"""
Module E0 - evaluation
======================

Module E0 evaluates the oemof results and calculates the KPI
- define dictionary entry for kpi matrix
- define dictionary entry for cost matrix
- store all results to matrix
"""

import logging

import oemof.solph as solph
import pandas as pd

import multi_vector_simulator.E1_process_results as E1
import multi_vector_simulator.E2_economics as E2
import multi_vector_simulator.E3_indicator_calculation as E3

import multi_vector_simulator.E4_verification as E4

from multi_vector_simulator.utils.constants import SOC

from multi_vector_simulator.utils.constants_json_strings import (
    UNIT,
    ENERGY_CONVERSION,
    ENERGY_CONSUMPTION,
    ENERGY_PRODUCTION,
    ENERGY_STORAGE,
    ENERGY_BUSSES,
    VALUE,
    SIMULATION_SETTINGS,
    ECONOMIC_DATA,
    INPUT_POWER,
    OUTPUT_POWER,
    STORAGE_CAPACITY,
    INFLOW_DIRECTION,
    OUTFLOW_DIRECTION,
    KPI,
    KPI_COST_MATRIX,
    KPI_SCALAR_MATRIX,
    KPI_SCALARS_DICT,
    OPTIMIZED_FLOWS,
    LABEL,
    INSTALLED_CAP,
    TIMESERIES_SOC,
    KPI_UNCOUPLED_DICT,
    CONSTRAINTS,
    MAXIMUM_EMISSIONS,
    NET_ZERO_ENERGY,
    MINIMAL_DEGREE_OF_AUTONOMY,
    MINIMAL_RENEWABLE_FACTOR,
    RENEWABLE_FACTOR,
    DEGREE_OF_AUTONOMY,
    FIX_COST,
    OPTIMIZED_ADD_CAP,
    LIFETIME_PRICE_DISPATCH,
    FLOW,
    COST_DISPATCH,
)

from multi_vector_simulator.utils.constants_output import (
    KPI_COST_MATRIX_ENTRIES,
    KPI_SCALAR_MATRIX_ENTRIES,
)


def evaluate_dict(dict_values, results_main, results_meta):
    """
    Processes all simulation outputs by evaluating oemof results, asset capacities and dispatch as well as all KPIs.

    Parameters
    ----------
    dict_values: dict
        simulation parameters

    results_main: DataFrame
        oemof simulation results as output by processing.results()

    results_meta: DataFrame
        oemof simulation meta information as output by processing.meta_results()

    Returns
    -------

    Notes
    -----

    Tested with:
    - test_E0.evaluation.test_evaluate_dict_append_new_fields()
    - test_E0.evaluation.test_evaluate_dict_important_fields_in_output_dict()
    - test_E0.evaluation.test_evaluate_dict_fields_values_in_output_dict_are_dataframes()
    """

    initalize_kpi(dict_values)

    bus_data = {}
    # Store all information related to busses in bus_data
    for bus in dict_values[ENERGY_BUSSES]:
        # Read all energy flows from busses
        bus_data.update({bus: solph.views.node(results_main, bus)})

    logging.info("Evaluating optimized capacities and dispatch.")
    # Evaluate timeseries and store to a large DataFrame for each bus:
    E1.get_timeseries_per_bus(dict_values, bus_data)

    # Store all information related to storages in bus_data, as storage capacity acts as a bus
    for storage in dict_values[ENERGY_STORAGE]:
        bus_data.update(
            {
                dict_values[ENERGY_STORAGE][storage][LABEL]: solph.views.node(
                    results_main, dict_values[ENERGY_STORAGE][storage][LABEL],
                )
            }
        )
        E1.get_storage_results(
            dict_values[SIMULATION_SETTINGS],
            bus_data[dict_values[ENERGY_STORAGE][storage][LABEL]],
            dict_values[ENERGY_STORAGE][storage],
        )

        for storage_item in [STORAGE_CAPACITY, INPUT_POWER, OUTPUT_POWER]:
            E2.get_costs(
                dict_values[ENERGY_STORAGE][storage][storage_item],
                dict_values[ECONOMIC_DATA],
            )

        E2.lcoe_assets(dict_values[ENERGY_STORAGE][storage], ENERGY_STORAGE)
        for storage_item in [STORAGE_CAPACITY, INPUT_POWER, OUTPUT_POWER]:
            store_result_matrix(
                dict_values[KPI], dict_values[ENERGY_STORAGE][storage][storage_item]
            )

        if (
            dict_values[ENERGY_STORAGE][storage][INFLOW_DIRECTION]
            in dict_values[OPTIMIZED_FLOWS].keys()
        ) or (
            dict_values[ENERGY_STORAGE][storage][OUTFLOW_DIRECTION]
            in dict_values[OPTIMIZED_FLOWS].keys()
        ):
            bus_name = dict_values[ENERGY_STORAGE][storage][INFLOW_DIRECTION]
            inflow_direction = dict_values[ENERGY_STORAGE][storage][INFLOW_DIRECTION]
            timeseries_name = (
                dict_values[ENERGY_STORAGE][storage][LABEL]
                + " ("
                + str(
                    round(
                        dict_values[ENERGY_STORAGE][storage][STORAGE_CAPACITY][
                            OPTIMIZED_ADD_CAP
                        ][VALUE]
                        + dict_values[ENERGY_STORAGE][storage][STORAGE_CAPACITY][
                            INSTALLED_CAP
                        ][VALUE],
                        1,
                    )
                )
                + dict_values[ENERGY_STORAGE][storage][STORAGE_CAPACITY][
                    OPTIMIZED_ADD_CAP
                ][UNIT]
                + f") {SOC}"
            )

            dict_values[OPTIMIZED_FLOWS][inflow_direction][
                timeseries_name
            ] = dict_values[ENERGY_STORAGE][storage][TIMESERIES_SOC]

    for group in [ENERGY_CONVERSION, ENERGY_PRODUCTION, ENERGY_CONSUMPTION]:
        for asset in dict_values[group]:
            E1.get_results(
                settings=dict_values[SIMULATION_SETTINGS],
                bus_data=bus_data,
                dict_asset=dict_values[group][asset],
                asset_group=group,
            )
            E2.get_costs(dict_values[group][asset], dict_values[ECONOMIC_DATA])
            E2.lcoe_assets(dict_values[group][asset], group)
            if group == ENERGY_PRODUCTION:
                E3.calculate_emissions_from_flow(dict_values[group][asset])
            store_result_matrix(dict_values[KPI], dict_values[group][asset])

    # Add fix project costs
    process_fixcost(dict_values)

    logging.info("Evaluating key performance indicators of the system")
    E3.all_totals(dict_values)
    E3.total_demand_and_excess_each_sector(dict_values)
    E3.add_total_feedin_electricity_equivalent(dict_values)
    E3.add_total_consumption_from_provider_electricity_equivalent(dict_values)
    E3.add_levelized_cost_of_energy_carriers(dict_values)
    E3.add_total_renewable_and_non_renewable_energy_origin(dict_values)
    E3.add_renewable_share_of_local_generation(dict_values)
    E3.add_renewable_factor(dict_values)
    # E3.add_degree_of_sector_coupling(dict_values) feature not finished
    E3.add_onsite_energy_fraction(dict_values)
    E3.add_onsite_energy_matching(dict_values)
    E3.add_degree_of_autonomy(dict_values)
    E3.add_degree_of_net_zero_energy(dict_values)
    E3.add_total_emissions(dict_values)
    E3.add_specific_emissions_per_electricity_equivalent(dict_values)

    # Tests and checks
    logging.info("Running validity checks.")

    if MINIMAL_RENEWABLE_FACTOR in dict_values[CONSTRAINTS]:
        E4.minimal_constraint_test(
            dict_values, MINIMAL_RENEWABLE_FACTOR, RENEWABLE_FACTOR
        )

    if MINIMAL_DEGREE_OF_AUTONOMY in dict_values[CONSTRAINTS]:
        E4.minimal_constraint_test(
            dict_values, MINIMAL_DEGREE_OF_AUTONOMY, DEGREE_OF_AUTONOMY
        )
    if MAXIMUM_EMISSIONS in dict_values[CONSTRAINTS]:
        E4.maximum_emissions_test(dict_values)

    if NET_ZERO_ENERGY in dict_values[CONSTRAINTS]:
        E4.net_zero_energy_constraint_test(dict_values)

    E4.detect_excessive_excess_generation_in_bus(dict_values)
    E4.verify_state_of_charge(dict_values)

    # Format KPI about energy generation into a pd.DataFrame
    dict_values[KPI][KPI_UNCOUPLED_DICT] = pd.DataFrame.from_dict(
        dict_values[KPI][KPI_UNCOUPLED_DICT], orient="index"
    )


def store_result_matrix(dict_kpi, dict_asset, fix_cost=False):
    """
    Storing results to vector and then result matrix for saving it in csv.
    Defined value types: Str, bool, None, dict (with key "VALUE"), else (int, float)

    Parameters
    ----------
    dict_kpi: dict
        dictionary with the two kpi groups (costs and scalars), which are pd.DF

    dict_asset: dict
        all information known for a specific asset

    fix_cost: Boolean
        If fix_cost is True, then no new row is added to KPI_SCALAR_MATRIX, as there are no KPI to update.
        Costs in KPI_COST_MATRIX however are added.

    Returns
    -------
    Updated dict_kpi DF, with new row of kpis of the specific asset.


    """

    round_to_comma = 5

    for kpi_storage in [KPI_COST_MATRIX, KPI_SCALAR_MATRIX]:
        if fix_cost == True and kpi_storage == KPI_SCALAR_MATRIX:
            pass
        else:
            asset_result_dict = {}
            for key in dict_kpi[kpi_storage].columns.values:
                # Check if called value is in oemof results -> Remember: check if pandas index has certain index: pd.object.index.contains(key)
                if key in dict_asset:
                    if isinstance(dict_asset[key], str):
                        asset_result_dict.update({key: dict_asset[key]})
                    elif isinstance(dict_asset[key], bool):
                        asset_result_dict.update({key: dict_asset[key]})
                    elif dict_asset[key] is None:
                        asset_result_dict.update({key: None})
                    elif isinstance(dict_asset[key], dict):
                        if VALUE in dict_asset[key].keys():
                            if dict_asset[key][VALUE] is not None:
                                asset_result_dict.update(
                                    {key: round(dict_asset[key][VALUE], round_to_comma)}
                                )
                    else:
                        asset_result_dict.update(
                            {key: round(dict_asset[key], round_to_comma)}
                        )

            asset_result_df = pd.DataFrame([asset_result_dict])

            dict_kpi.update(
                {
                    kpi_storage: pd.concat(
                        [dict_kpi[kpi_storage], asset_result_df], ignore_index=True
                    )
                }
            )


def initalize_kpi(dict_values):
    r"""
    Adds basic structure of KPI to dict_values to gather them later on.

    Parameters
    ----------
    dict_values: dict
        All simulation data, but without any results

    Returns
    -------
    Updated dict_values with KPI structure, made up from KPI_COST_MATRIX, KPI_SCALAR_MATRIX and KPI_SCALARS_DICT.

    """
    dict_values.update(
        {
            KPI: {
                KPI_COST_MATRIX: pd.DataFrame(columns=KPI_COST_MATRIX_ENTRIES),
                KPI_SCALAR_MATRIX: pd.DataFrame(columns=KPI_SCALAR_MATRIX_ENTRIES),
                KPI_SCALARS_DICT: {},
            }
        }
    )


def process_fixcost(dict_values):
    r"""
    Adds fix costs of the project to the economic evaluation of the energy system.

    Parameters
    ----------
    dict_values: dict
        All simulation data with inputs and results of the assets

    Returns
    -------
    Updated dict_values with costs attributed in dict values also appended to the dict_values[KPI] (scalar results)

    Notes
    -----

    Function is tested with:
    - test_E0_evaluation.test_process_fixcost()
    """
    for asset in dict_values[FIX_COST]:
        # Add parameters that are needed for E2.get_costs()
        dict_values[FIX_COST][asset].update(
            {
                OPTIMIZED_ADD_CAP: {VALUE: 1},
                INSTALLED_CAP: {VALUE: 0},
                LIFETIME_PRICE_DISPATCH: {VALUE: 0},
                FLOW: pd.Series([0, 0]),
            }
        )

        E2.get_costs(dict_values[FIX_COST][asset], dict_values[ECONOMIC_DATA])
        # Remove all parameters that were added before and the KPI that do not apply
        for key in [
            OPTIMIZED_ADD_CAP,
            LIFETIME_PRICE_DISPATCH,
            INSTALLED_CAP,
            FLOW,
            COST_DISPATCH,
        ]:
            dict_values[FIX_COST][asset].pop(key)
        store_result_matrix(
            dict_values[KPI], dict_values[FIX_COST][asset], fix_cost=True
        )



"""
Module D1 - Oemof components
============================

Module D1 includes all functions that are required to build an oemof model with adaptable components.

- add transformer objects (fix, to be optimized)
- add source objects (fix, to be optimized, dispatchable, non-dispatchable)
- add sink objects (fix, to be optimized, dispatchable, non-dispatchable)
- add storage objects (fix, to be optimized)
- add multiple input/output busses if required for each of the assets
- add oemof component parameters as scalar or time series values

"""

import logging

import oemof.solph as solph

from multi_vector_simulator.utils.constants_json_strings import (
    VALUE,
    UNIT,
    LABEL,
    DISPATCH_PRICE,
    AVAILABILITY_DISPATCH,
    OPTIMIZE_CAP,
    INSTALLED_CAP,
    INSTALLED_CAP_NORMALIZED,
    EFFICIENCY,
    ENERGY_VECTOR,
    INPUT_POWER,
    OUTPUT_POWER,
    C_RATE,
    SOC_INITIAL,
    SOC_MAX,
    SOC_MIN,
    THERM_LOSSES_REL,
    THERM_LOSSES_ABS,
    STORAGE_CAPACITY,
    TIMESERIES,
    TIMESERIES_NORMALIZED,
    TIMESERIES_PEAK,
    INFLOW_DIRECTION,
    OUTFLOW_DIRECTION,
    SIMULATION_ANNUITY,
    MAXIMUM_ADD_CAP,
    MAXIMUM_ADD_CAP_NORMALIZED,
    DISPATCHABILITY,
    OEMOF_ASSET_TYPE,
    OEMOF_GEN_STORAGE,
    OEMOF_SINK,
    OEMOF_SOURCE,
    OEMOF_TRANSFORMER,
    OEMOF_BUSSES,
    OEMOF_ExtractionTurbineCHP,
    EMISSION_FACTOR,
    BETA,
)
from multi_vector_simulator.utils.helpers import get_item_if_list, get_length_if_list
from multi_vector_simulator.utils.exceptions import (
    MissingParameterError,
    WrongParameterFormatError,
)


def check_list_parameters_transformers_single_input_single_output(dict_asset):
    parameters_defined_as_list = []
    for parameter in [DISPATCH_PRICE, EFFICIENCY]:
        if get_length_if_list(dict_asset[parameter][VALUE]) != 0:
            parameters_defined_as_list.append(parameter)

    if parameters_defined_as_list:
        parameters_defined_as_list = ", ".join(parameters_defined_as_list)
        missing_dispatch_prices_or_efficiencies = (
            f"You defined multiple values for parameter(s) '{parameters_defined_as_list}'"
            f" although you you have one input and one output for"
            f" the conversion asset named '{dict_asset[LABEL]}', please provide only scalars"
            f" or define more input/output busses"
        )
        logging.error(missing_dispatch_prices_or_efficiencies)
        raise ValueError(missing_dispatch_prices_or_efficiencies)


def transformer(model, dict_asset, **kwargs):
    r"""
    Defines a transformer component specified in `dict_asset`.

    Depending on the 'value' of 'optimizeCap' in `dict_asset` the transformer
    is defined with a fixed capacity or a capacity to be optimized.
    The transformer has multiple or single input or output busses depending on
    the types of keys 'inflow_direction' and 'outflow_direction' in `dict_asset`.

    Parameters
    ----------
    model : oemof.solph.network.EnergySystem object
        See the oemof documentation for more information.
    dict_asset : dict
        Contains information about the transformer like (not exhaustive):
        efficiency, installed capacity ('installedCap'), information on the
        busses the transformer is connected to ('inflow_direction',
        'outflow_direction').

    Other Parameters
    ----------------
    busses : dict
    sinks : dict, optional
    sources : dict, optional
    transformers : dict
    storages : dict, optional

    Notes
    -----
    The transformer has either multiple input or multiple output busses.

    The following functions are used for defining the transformer:
    * :py:func:`~.transformer_constant_efficiency_fix`
    * :py:func:`~.transformer_constant_efficiency_optimize`

    Tested with:
    - test_transformer_optimize_cap_single_busses()
    - test_transformer_optimize_cap_multiple_input_busses()
    - test_transformer_optimize_cap_multiple_output_busses()
    - test_transformer_fix_cap_single_busses()
    - test_transformer_fix_cap_multiple_input_busses()
    - test_transformer_fix_cap_multiple_output_busses()

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with transformer object.

    """
    check_optimize_cap(
        model,
        dict_asset,
        func_constant=transformer_constant_efficiency_fix,
        func_optimize=transformer_constant_efficiency_optimize,
        **kwargs,
    )


def chp(model, dict_asset, **kwargs):
    r"""
    Defines a chp component specified in `dict_asset`.

    Depending on the 'value' of 'optimizeCap' in `dict_asset` the chp
    is defined with a fixed capacity or a capacity to be optimized.
    The chp has single input and multiple output busses.

    Parameters
    ----------
    model : oemof.solph.network.EnergySystem object
        See the oemof documentation for more information.
    dict_asset : dict
        Contains information about the chp like (not exhaustive):
        efficiency, installed capacity ('installedCap'), information on the
        busses the chp is connected to ('inflow_direction',
        'outflow_direction'), beta coefficient.

    Other Parameters
    ----------------
    busses : dict
    sinks : dict, optional
    sources : dict, optional
    transformers : dict
    storages : dict, optional
    extractionTurbineCHP: dict, optional

    Notes
    -----
    The transformer has either multiple input or multiple output busses.

    The following functions are used for defining the chp:
    * :py:func:`~.chp_fix`
    * :py:func:`~.chp_optimize` for investment optimization

    Tested with:
    - test_chp_fix_cap()
    - test_chp_optimize_cap()
    - test_chp_missing_beta()
    - test_chp_wrong_beta_formatting()
    - test_chp_wrong_efficiency_formatting()
    - test_chp_wrong_outflow_bus_energy_vector()

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with chp object.

    """
    if BETA in dict_asset:
        beta = dict_asset[BETA]
        if isinstance(beta, dict) is False:
            raise WrongParameterFormatError(
                f"For the conversion asset named '{dict_asset[LABEL]}' of type {OEMOF_ExtractionTurbineCHP}, "
                f"the {BETA} parameter should have the following format {{ '{VALUE}': ..., '{UNIT}': ... }}"
            )
        else:
            beta = beta[VALUE]
        if 0 <= beta <= 1:
            pass
        else:
            raise ValueError("beta should be a number between 0 and 1.")
    else:
        raise MissingParameterError("No beta for extraction turbine chp specified.")

    if isinstance(dict_asset[EFFICIENCY][VALUE], list) is False:
        missing_efficiencies = (
            f"For the conversion asset named '{dict_asset[LABEL]}' of type {OEMOF_ExtractionTurbineCHP} "
            f"you must provide exactly 2 values for the parameter '{EFFICIENCY}'."
        )
        logging.error(missing_efficiencies)
        raise WrongParameterFormatError(missing_efficiencies)

    busses_energy_vectors = [
        kwargs[OEMOF_BUSSES][b].energy_vector for b in dict_asset[OUTFLOW_DIRECTION]
    ]
    if (
        "Heat" not in busses_energy_vectors
        or "Electricity" not in busses_energy_vectors
    ):
        mapping_busses = [
            f"'{v}' (from '{k}')"
            for k, v in zip(dict_asset[OUTFLOW_DIRECTION], busses_energy_vectors)
        ]
        wrong_output_energy_vectors = (
            f"For the conversion asset named '{dict_asset[LABEL]}' of type {OEMOF_ExtractionTurbineCHP} "
            f"you must provide 1 output bus for energy vector 'Heat' and one for 'Electricity'. You provided "
            f"{' and '.join(mapping_busses)}"
        )
        logging.error(wrong_output_energy_vectors)
        raise WrongParameterFormatError(wrong_output_energy_vectors)

    check_optimize_cap(
        model, dict_asset, func_constant=chp_fix, func_optimize=chp_optimize, **kwargs
    )


def storage(model, dict_asset, **kwargs):
    r"""
    Defines a storage component specified in `dict_asset`.

    Depending on the 'value' of 'optimizeCap' in `dict_asset` the storage
    is defined with a fixed capacity or a capacity to be optimized.

    Parameters
    ----------
    model : oemof.solph.network.EnergySystem object
        See the oemof documentation for more information.
    dict_asset : dict
        Contains information about the storage like (not exhaustive):
        efficiency, installed capacity ('installedCap'), information on the
        busses the storage is connected to ('inflow_direction',
        'outflow_direction'),

    Other Parameters
    ----------------
    busses : dict
    sinks : dict, optional
    sources : dict, optional
    transformers : dict, optional
    storages : dict

    Notes
    -----
    The following functions are used for defining the storage:
    * :py:func:`~.storage_fix`
    * :py:func:`~.storage_optimize`

    Tested with:
    - test_storage_optimize()
    - test_storage_fix()

    """
    check_optimize_cap(
        model,
        dict_asset,
        func_constant=storage_fix,
        func_optimize=storage_optimize,
        **kwargs,
    )


def sink(model, dict_asset, **kwargs):
    r"""
    Defines a sink component specified in `dict_asset`.

    Depending on the 'value' of 'optimizeCap' in `dict_asset` the sink
    is defined with a fixed capacity or a capacity to be optimized. If a time
    series is provided for the sink (key 'timeseries' in `dict_asset`) it is
    defined as a non dispatchable sink, otherwise as dispatchable sink.
    The sink has multiple or a single input bus depending on the type of the
    key 'inflow_direction' in `dict_asset`.

    Parameters
    ----------
    model : oemof.solph.network.EnergySystem object
        See the oemof documentation for more information.
    dict_asset : dict
        Contains information about the storage like (not exhaustive):
        efficiency, installed capacity ('installedCap'), information on the
        busses the sink is connected to ('inflow_direction'),

    Other Parameters
    ----------------
    busses : dict
    sinks : dict
    sources : dict, optional
    transformers : dict, optional
    storages : dict, optional

    Notes
    -----
    The following functions are used for defining the sink:
    * :py:func:`~.sink_non_dispatchable`
    * :py:func:`~.sink_dispatchable`

    Tested with:
    - test_sink_non_dispatchable_single_input_bus()
    - test_sink_non_dispatchable_multiple_input_busses()
    - test_sink_dispatchable_single_input_bus()
    - test_sink_dispatchable_multiple_input_busses()

    """
    if TIMESERIES in dict_asset:
        sink_non_dispatchable(model, dict_asset, **kwargs)
    else:
        sink_dispatchable_optimize(model, dict_asset, **kwargs)


def source(model, dict_asset, **kwargs):
    r"""
    Defines a source component specified in `dict_asset`.

    Depending on the 'value' of 'optimizeCap' in `dict_asset` the source
    is defined with a fixed capacity or a capacity to be optimized. If a time
    series is provided for the source (key 'timeseries' in `dict_asset`) it is
    defined as a non dispatchable source, otherwise as dispatchable source.
    The source has multiple or a single output bus depending on the type of the
    key 'inflow_direction' in `dict_asset`.

    Parameters
    ----------
    model : oemof.solph.network.EnergySystem object
        See the oemof documentation for more information.
    dict_asset : dict
        Contains information about the storage like (not exhaustive):
        efficiency, installed capacity ('installedCap'), information on the
        busses the sink is connected to ('inflow_direction'),

    Other Parameters
    ----------------
    busses : dict
    sinks : dict
    sources : dict, optional
    transformers : dict, optional
    storages : dict, optional

    TODOS
    ^^^^^
    * We should actually not allow multiple output busses, probably - because a pv would then
    feed in twice as much as solar_gen_specific for example, see issue #121

    Notes
    -----
    The following functions are used for defining the source:
    * :py:func:`~.source_dispatchable_fix`
    * :py:func:`~.source_dispatchable_optimize`
    * :py:func:`~.source_non_dispatchable_fix`
    * :py:func:`~.source_non_dispatchable_optimize`

    Tested with:
    - test_source_non_dispatchable_optimize()
    - test_source_non_dispatchable_fix()
    - test_source_dispatchable_optimize_normalized_timeseries()
    - test_source_dispatchable_optimize_timeseries_not_normalized_timeseries()
    - test_source_dispatchable_fix_normalized_timeseries()
    - test_source_dispatchable_fix_timeseries_not_normalized_timeseries()
    """
    if DISPATCHABILITY in dict_asset and dict_asset[DISPATCHABILITY] is True:
        check_optimize_cap(
            model,
            dict_asset,
            func_constant=source_dispatchable_fix,
            func_optimize=source_dispatchable_optimize,
            **kwargs,
        )

    else:
        check_optimize_cap(
            model,
            dict_asset,
            func_constant=source_non_dispatchable_fix,
            func_optimize=source_non_dispatchable_optimize,
            **kwargs,
        )


def check_optimize_cap(model, dict_asset, func_constant, func_optimize, **kwargs):
    r"""
    Defines a component specified in `dict_asset` with fixed capacity or capacity to be optimized.

    Parameters
    ----------
    model : oemof.solph.network.EnergySystem object
        See the oemof documentation for more information.
    dict_asset : dict
        Contains information about the asset like (not exhaustive):
        efficiency, installed capacity ('installedCap'), information on the
        busses the asset is connected to (f.e. 'inflow_direction',
        'outflow_direction').
    func_constant : func
        Function to be applied if optimization not intended.
    func_optimize : func
        Function to be applied if optimization not intended.

    Other Parameters
    ----------------
    Required are `busses` and a dictionary belonging to the respective oemof
    type of the asset.

    busses : dict, optional
    sinks : dict, optional
    sources : dict, optional
    transformers : dict, optional
    storages : dict, optional

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the component object.

    TODOS
    ^^^^^
    Might be possible to drop non invest optimization in favour of invest optimization if max_capactiy
    attributes ie. are set to 0 for fix (but less beautiful, and in case of generator even blocks nonconvex opt.).

    Notes
    -----
    Tested with:
    - test_check_optimize_cap_raise_error()

    """
    if dict_asset[OPTIMIZE_CAP][VALUE] is False:
        func_constant(model, dict_asset, **kwargs)
        if dict_asset[OEMOF_ASSET_TYPE] != OEMOF_SOURCE:
            logging.debug(
                "Added: %s %s (fixed capacity)",
                dict_asset[OEMOF_ASSET_TYPE].capitalize(),
                dict_asset[LABEL],
            )

    elif dict_asset[OPTIMIZE_CAP][VALUE] is True:
        func_optimize(model, dict_asset, **kwargs)
        if dict_asset[OEMOF_ASSET_TYPE] != OEMOF_SOURCE:
            logging.debug(
                "Added: %s %s (capacity to be optimized)",
                dict_asset[OEMOF_ASSET_TYPE].capitalize(),
                dict_asset[LABEL],
            )
    else:
        raise ValueError(
            f"Input error! '{OPTIMIZE_CAP}' of asset {dict_asset[LABEL]}\n should be True/False but is {dict_asset[OPTIMIZE_CAP][VALUE]}."
        )


class CustomBus(solph.Bus):
    def __init__(self, *args, **kwargs):
        ev = kwargs.pop("energy_vector", None)  # change to ENERGY_VECTOR
        super(CustomBus, self).__init__(*args, **kwargs)
        self.energy_vector = ev


def bus(model, name, **kwargs):
    r"""
    Adds bus `name` to `model` and to 'busses' in `kwargs`.

    Notes
    -----
    Tested with:
    - test_bus_add_to_empty_dict()
    - test_bus_add_to_not_empty_dict()

    """
    logging.debug(f"Added: Bus {name}")
    energy_vector = kwargs.get("energy_vector", None)  # change to ENERGY_VECTOR
    bus = CustomBus(label=name, energy_vector=energy_vector)
    kwargs[OEMOF_BUSSES].update({name: bus})
    model.add(bus)


def transformer_constant_efficiency_fix(model, dict_asset, **kwargs):
    r"""
    Defines a transformer with constant efficiency and fixed capacity.

    See :py:func:`~.transformer` for more information, including parameters.

    Notes
    -----
    Tested with:
    - test_transformer_fix_cap_single_busses()
    - test_transformer_fix_cap_multiple_input_busses()
    - test_transformer_fix_cap_multiple_output_busses()

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the transformer object.

    """

    missing_dispatch_prices_or_efficiencies = None

    # check if the transformer has multiple input or multiple output busses
    if isinstance(dict_asset[INFLOW_DIRECTION], list) or isinstance(
        dict_asset[OUTFLOW_DIRECTION], list
    ):
        if isinstance(dict_asset[INFLOW_DIRECTION], list) and isinstance(
            dict_asset[OUTFLOW_DIRECTION], str
        ):
            # multiple inputs and single output
            inputs = {}

            num_inputs = len(dict_asset[INFLOW_DIRECTION])
            inputs_names = ", ".join([f"'{n}'" for n in dict_asset[INFLOW_DIRECTION]])
            if get_length_if_list(dict_asset[EFFICIENCY][VALUE]) != num_inputs:
                missing_dispatch_prices_or_efficiencies = (
                    f"You defined multiple values for parameter '{INFLOW_DIRECTION}' "
                    f"({inputs_names}) of the conversion asset named '{dict_asset[LABEL]}'. "
                    f"You must also provide exactly {num_inputs} values for the parameter '{EFFICIENCY}'."
                )
                logging.error(missing_dispatch_prices_or_efficiencies)
                raise ValueError(missing_dispatch_prices_or_efficiencies)

            if get_length_if_list(dict_asset[DISPATCH_PRICE][VALUE]) == 0:
                # only one dispatch price provided --> it will be ignored
                warning_msg = (
                    f"You defined multiple values for parameter '{INFLOW_DIRECTION}' "
                    f"({inputs_names}) of the conversion asset named '{dict_asset[LABEL]}'. "
                    f"You can also provide exactly {num_inputs} values for the parameter '{DISPATCH_PRICE}'."
                    f"You did only provide one value, so we will ignore it"
                )
                logging.warning(warning_msg)
                for i, bus in enumerate(dict_asset[INFLOW_DIRECTION]):
                    inputs[kwargs[OEMOF_BUSSES][bus]] = solph.Flow()
            else:
                for i, bus in enumerate(dict_asset[INFLOW_DIRECTION]):
                    inputs[kwargs[OEMOF_BUSSES][bus]] = solph.Flow(
                        variable_costs=get_item_if_list(
                            dict_asset[DISPATCH_PRICE][VALUE], i
                        )
                    )

            outputs = {
                kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
                    nominal_value=dict_asset[INSTALLED_CAP][VALUE]
                )
            }
            efficiencies = {}
            for i, efficiency in enumerate(dict_asset[EFFICIENCY][VALUE]):
                efficiencies[
                    kwargs[OEMOF_BUSSES][dict_asset[INFLOW_DIRECTION][i]]
                ] = efficiency

        elif isinstance(dict_asset[INFLOW_DIRECTION], str) and isinstance(
            dict_asset[OUTFLOW_DIRECTION], list
        ):
            # single input and multiple outputs
            num_outputs = len(dict_asset[OUTFLOW_DIRECTION])
            if get_length_if_list(dict_asset[EFFICIENCY][VALUE]) != num_outputs:
                outputs_names = ", ".join(
                    [f"'{n}'" for n in dict_asset[OUTFLOW_DIRECTION]]
                )
                missing_dispatch_prices_or_efficiencies = (
                    f"You defined multiple values for parameter '{OUTFLOW_DIRECTION}' "
                    f"({outputs_names}) of the conversion asset named '{dict_asset[LABEL]}'. "
                    f"You must also provide exactly {num_outputs} values for the parameters "
                    f"'{EFFICIENCY}' and you can do so for the parameter '{DISPATCH_PRICE}'."
                )
                logging.error(missing_dispatch_prices_or_efficiencies)
                raise ValueError(missing_dispatch_prices_or_efficiencies)

            inputs = {kwargs[OEMOF_BUSSES][dict_asset[INFLOW_DIRECTION]]: solph.Flow()}
            outputs = {}
            for i, bus in enumerate(dict_asset[OUTFLOW_DIRECTION]):
                outputs[kwargs[OEMOF_BUSSES][bus]] = solph.Flow(
                    nominal_value=get_item_if_list(dict_asset[INSTALLED_CAP][VALUE], i),
                    variable_costs=get_item_if_list(
                        dict_asset[DISPATCH_PRICE][VALUE], i
                    ),
                )

            efficiencies = {}
            for i, efficiency in enumerate(dict_asset[EFFICIENCY][VALUE]):
                efficiencies[
                    kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION][i]]
                ] = efficiency

        else:
            # multiple inputs and multiple outputs
            inputs_names = ", ".join([f"'{n}'" for n in dict_asset[INFLOW_DIRECTION]])
            outputs_names = ", ".join([f"'{n}'" for n in dict_asset[OUTFLOW_DIRECTION]])
            missing_dispatch_prices_or_efficiencies = (
                f"You defined multiple values for parameter '{INFLOW_DIRECTION}'"
                f" ({inputs_names}) as well as for parameter '{OUTFLOW_DIRECTION}' ({outputs_names})"
                f" of the conversion asset named '{dict_asset[LABEL]}', this is not supported"
                f" at the moment."
            )
            logging.error(missing_dispatch_prices_or_efficiencies)
            raise ValueError(missing_dispatch_prices_or_efficiencies)
    else:
        # single input and single output

        check_list_parameters_transformers_single_input_single_output(dict_asset)

        inputs = {kwargs[OEMOF_BUSSES][dict_asset[INFLOW_DIRECTION]]: solph.Flow()}
        outputs = {
            kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
                nominal_value=dict_asset[INSTALLED_CAP][VALUE],
                variable_costs=dict_asset[DISPATCH_PRICE][VALUE],
            )
        }
        efficiencies = {
            kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: dict_asset[EFFICIENCY][
                VALUE
            ]
        }

    if missing_dispatch_prices_or_efficiencies is None:
        t = solph.Transformer(
            label=dict_asset[LABEL],
            inputs=inputs,
            outputs=outputs,
            conversion_factors=efficiencies,
        )

        model.add(t)
        kwargs[OEMOF_TRANSFORMER].update({dict_asset[LABEL]: t})


def transformer_constant_efficiency_optimize(model, dict_asset, **kwargs):
    r"""
    Defines a transformer with constant efficiency and a capacity to be optimized.

    See :py:func:`~.transformer` for more information, including parameters.

    Notes
    -----
    Tested with:
    - test_transformer_optimize_cap_single_busses()
    - test_transformer_optimize_cap_multiple_input_busses()
    - test_transformer_optimize_cap_multiple_output_busses()

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the transformer object.

    """
    missing_dispatch_prices_or_efficiencies = None

    # check if the transformer has multiple input or multiple output busses
    # the investment object is always in the output bus
    if isinstance(dict_asset[INFLOW_DIRECTION], list) or isinstance(
        dict_asset[OUTFLOW_DIRECTION], list
    ):
        if isinstance(dict_asset[INFLOW_DIRECTION], list) and isinstance(
            dict_asset[OUTFLOW_DIRECTION], str
        ):
            # multiple inputs and single output
            num_inputs = len(dict_asset[INFLOW_DIRECTION])
            if get_length_if_list(dict_asset[EFFICIENCY][VALUE]) != num_inputs:
                inputs_names = ", ".join(
                    [f"'{n}'" for n in dict_asset[INFLOW_DIRECTION]]
                )
                missing_dispatch_prices_or_efficiencies = (
                    f"You defined multiple values for parameter '{INFLOW_DIRECTION}' "
                    f"({inputs_names}) of the conversion asset named '{dict_asset[LABEL]}'. "
                    f"You must also provide exactly {num_inputs} values for the parameter "
                    f"'{EFFICIENCY}' and you can do so for the parameter '{DISPATCH_PRICE}'."
                )
                logging.error(missing_dispatch_prices_or_efficiencies)
                raise ValueError(missing_dispatch_prices_or_efficiencies)

            inputs = {}
            for i, bus in enumerate(dict_asset[INFLOW_DIRECTION]):
                inputs[kwargs[OEMOF_BUSSES][bus]] = solph.Flow(
                    variable_costs=get_item_if_list(
                        dict_asset[DISPATCH_PRICE][VALUE], i
                    )
                )

            outputs = {
                kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
                    investment=solph.Investment(
                        ep_costs=dict_asset[SIMULATION_ANNUITY][VALUE],
                        maximum=dict_asset[MAXIMUM_ADD_CAP][VALUE],
                        existing=dict_asset[INSTALLED_CAP][VALUE],
                    )
                )
            }

            efficiencies = {}
            for i, efficiency in enumerate(dict_asset[EFFICIENCY][VALUE]):
                efficiencies[
                    kwargs[OEMOF_BUSSES][dict_asset[INFLOW_DIRECTION][i]]
                ] = efficiency

        elif isinstance(dict_asset[INFLOW_DIRECTION], str) and isinstance(
            dict_asset[OUTFLOW_DIRECTION], list
        ):
            # single input and multiple outputs
            num_outputs = len(dict_asset[OUTFLOW_DIRECTION])
            if get_length_if_list(dict_asset[EFFICIENCY][VALUE]) != num_outputs:
                outputs_names = ", ".join(
                    [f"'{n}'" for n in dict_asset[OUTFLOW_DIRECTION]]
                )
                missing_dispatch_prices_or_efficiencies = (
                    f"You defined multiple values for parameter '{OUTFLOW_DIRECTION}' "
                    f"({outputs_names}) of the conversion asset named '{dict_asset[LABEL]}'. "
                    f"You must also provide exactly {num_outputs} values for the parameter "
                    f"'{EFFICIENCY}' and you can do so for the parameter '{DISPATCH_PRICE}'."
                )
                logging.error(missing_dispatch_prices_or_efficiencies)
                raise ValueError(missing_dispatch_prices_or_efficiencies)

            inputs = {kwargs[OEMOF_BUSSES][dict_asset[INFLOW_DIRECTION]]: solph.Flow()}
            outputs = {}
            efficiencies = {}

            for i, (bus, efficiency) in enumerate(
                zip(dict_asset[OUTFLOW_DIRECTION], dict_asset[EFFICIENCY][VALUE])
            ):
                if i == 0:
                    investment_params = dict(
                        investment=solph.Investment(
                            ep_costs=dict_asset[SIMULATION_ANNUITY][VALUE],
                            maximum=dict_asset[MAXIMUM_ADD_CAP][VALUE],
                            existing=dict_asset[INSTALLED_CAP][VALUE],
                        )
                    )
                else:
                    investment_params = {}
                outputs[kwargs[OEMOF_BUSSES][bus]] = solph.Flow(**investment_params)
                efficiencies[kwargs[OEMOF_BUSSES][bus]] = efficiency
        else:
            # multiple inputs and multiple outputs
            inputs_names = ", ".join([f"'{n}'" for n in dict_asset[INFLOW_DIRECTION]])
            outputs_names = ", ".join([f"'{n}'" for n in dict_asset[OUTFLOW_DIRECTION]])
            missing_dispatch_prices_or_efficiencies = (
                f"You defined multiple values for parameter '{INFLOW_DIRECTION}'"
                f" ({inputs_names}) as well as for parameter '{OUTFLOW_DIRECTION}' ({outputs_names})"
                f" of the conversion asset named '{dict_asset[LABEL]}', this is not supported"
                f" at the moment."
            )
            logging.error(missing_dispatch_prices_or_efficiencies)
            raise ValueError(missing_dispatch_prices_or_efficiencies)
    else:
        check_list_parameters_transformers_single_input_single_output(dict_asset)

        # single input and single output
        inputs = {kwargs[OEMOF_BUSSES][dict_asset[INFLOW_DIRECTION]]: solph.Flow()}
        if AVAILABILITY_DISPATCH in dict_asset.keys():
            # This key is only present in DSO peak demand pricing transformers.
            outputs = {
                kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
                    investment=solph.Investment(
                        ep_costs=dict_asset[SIMULATION_ANNUITY][VALUE],
                        maximum=dict_asset[MAXIMUM_ADD_CAP][VALUE],
                        existing=dict_asset[INSTALLED_CAP][VALUE],
                    ),
                    variable_costs=dict_asset[DISPATCH_PRICE][VALUE],
                    max=dict_asset[AVAILABILITY_DISPATCH].values,
                )
            }
        else:
            outputs = {
                kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
                    investment=solph.Investment(
                        ep_costs=dict_asset[SIMULATION_ANNUITY][VALUE],
                        maximum=dict_asset[MAXIMUM_ADD_CAP][VALUE],
                        existing=dict_asset[INSTALLED_CAP][VALUE],
                    ),
                    variable_costs=dict_asset[DISPATCH_PRICE][VALUE],
                )
            }

        efficiencies = {
            kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: dict_asset[EFFICIENCY][
                VALUE
            ]
        }

    if missing_dispatch_prices_or_efficiencies is None:
        t = solph.Transformer(
            label=dict_asset[LABEL],
            inputs=inputs,
            outputs=outputs,
            conversion_factors=efficiencies,
        )

        model.add(t)
        kwargs[OEMOF_TRANSFORMER].update({dict_asset[LABEL]: t})


def storage_fix(model, dict_asset, **kwargs):
    r"""
    Defines a storage with a fixed capacity.

    See :py:func:`~.storage` for more information, including parameters.

    Notes
    -----
    Tested with:
    - test_storage_fix()

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the storage object.

    """
    storage = solph.components.GenericStorage(
        label=dict_asset[LABEL],
        nominal_storage_capacity=dict_asset[STORAGE_CAPACITY][INSTALLED_CAP][VALUE], #THERMAL --> yes
        inputs={
            kwargs[OEMOF_BUSSES][dict_asset[INFLOW_DIRECTION]]: solph.Flow(
                nominal_value=dict_asset[INPUT_POWER][INSTALLED_CAP][
                    VALUE
                ],  # limited through installed capacity, NOT c-rate
                # might be too much
                variable_costs=dict_asset[INPUT_POWER][DISPATCH_PRICE][VALUE],
            )
        },  # maximum charge possible in one timestep
        outputs={
            kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
                nominal_value=dict_asset[OUTPUT_POWER][INSTALLED_CAP][
                    VALUE
                ],  # limited through installed capacity, NOT c-rate #todo actually, if we only have a lithium battery... crate should suffice? i mean, with crate fixed AND fixed power, this is defined two times
                variable_costs=dict_asset[OUTPUT_POWER][DISPATCH_PRICE][VALUE],
            )
        },  # maximum discharge possible in one timestep
        loss_rate=1 - dict_asset[STORAGE_CAPACITY][EFFICIENCY][VALUE],  # from timestep to timestep #THERMAL
        fixed_losses_absolute=dict_asset[STORAGE_CAPACITY][THERM_LOSSES_ABS][VALUE], #THERMAL
        fixed_losses_relative=dict_asset[STORAGE_CAPACITY][THERM_LOSSES_REL][VALUE], #THERMAL
        min_storage_level=dict_asset[STORAGE_CAPACITY][SOC_MIN][VALUE],
        max_storage_level=dict_asset[STORAGE_CAPACITY][SOC_MAX][VALUE],
        initial_storage_level=dict_asset[STORAGE_CAPACITY][SOC_INITIAL][
            VALUE
        ],  # in terms of SOC
        inflow_conversion_factor=dict_asset[INPUT_POWER][EFFICIENCY][
            VALUE
        ],  # storing efficiency
        outflow_conversion_factor=dict_asset[OUTPUT_POWER][EFFICIENCY][VALUE],
    )  # efficiency of discharge
    model.add(storage)
    kwargs[OEMOF_GEN_STORAGE].update({dict_asset[LABEL]: storage})


def storage_optimize(model, dict_asset, **kwargs):
    r"""
    Defines a storage with a capacity to be optimized.

    See :py:func:`~.storage` for more information, including parameters.

    Notes
    -----
    Tested with:
    - test_storage_optimize()
    - test_storage_optimize_investment_minimum_0_float()
    - test_storage_optimize_investment_minimum_0_time_series()
    - test_storage_optimize_investment_minimum_1_rel_float()
    - test_storage_optimize_investment_minimum_1_abs_float()
    - test_storage_optimize_investment_minimum_1_rel_times_series()
    - test_storage_optimize_investment_minimum_1_abs_times_series()

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the storage object.

    """
    # investment.minimum for an InvestmentStorage is 0 as default
    minimum = 0

    # Set investment.minimum to 1 if
    # non-zero fixed_thermal_losses_relative or fixed_thermal_losses_absolute exist as
    for losses in [THERM_LOSSES_REL, THERM_LOSSES_ABS]:
        # 1. float or
        try:
            float(dict_asset[STORAGE_CAPACITY][losses][VALUE])
            if dict_asset[STORAGE_CAPACITY][losses][VALUE] != 0:
                minimum = 1
        # 2. time series
        except TypeError:
            if sum(dict_asset[STORAGE_CAPACITY][losses][VALUE]) != 0:
                minimum = 1

    storage = solph.components.GenericStorage(
        label=dict_asset[LABEL],
        investment=solph.Investment(
            ep_costs=dict_asset[STORAGE_CAPACITY][SIMULATION_ANNUITY][VALUE],
            minimum=minimum,
            maximum=dict_asset[STORAGE_CAPACITY][MAXIMUM_ADD_CAP][VALUE],
            existing=dict_asset[STORAGE_CAPACITY][INSTALLED_CAP][VALUE],
        ),
        inputs={
            kwargs[OEMOF_BUSSES][dict_asset[INFLOW_DIRECTION]]: solph.Flow(
                investment=solph.Investment(
                    ep_costs=dict_asset[INPUT_POWER][SIMULATION_ANNUITY][VALUE],
                    maximum=dict_asset[INPUT_POWER][MAXIMUM_ADD_CAP][VALUE],
                    existing=dict_asset[INPUT_POWER][INSTALLED_CAP][
                        VALUE
                    ],  # todo: `existing needed here?`
                ),
                variable_costs=dict_asset[INPUT_POWER][DISPATCH_PRICE][VALUE],
            )
        },  # maximum charge power
        outputs={
            kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
                investment=solph.Investment(
                    ep_costs=dict_asset[OUTPUT_POWER][SIMULATION_ANNUITY][VALUE],
                    maximum=dict_asset[OUTPUT_POWER][MAXIMUM_ADD_CAP][VALUE],
                    existing=dict_asset[OUTPUT_POWER][INSTALLED_CAP][
                        VALUE
                    ],  # todo: `existing needed here?`
                ),
                variable_costs=dict_asset[OUTPUT_POWER][DISPATCH_PRICE][VALUE],
            )
        },  # maximum discharge power
        loss_rate=1
        - dict_asset[STORAGE_CAPACITY][EFFICIENCY][VALUE],  # from timestep to timestep
        fixed_losses_absolute=dict_asset[STORAGE_CAPACITY][THERM_LOSSES_ABS][VALUE],
        fixed_losses_relative=dict_asset[STORAGE_CAPACITY][THERM_LOSSES_REL][VALUE],
        min_storage_level=dict_asset[STORAGE_CAPACITY][SOC_MIN][VALUE],
        max_storage_level=dict_asset[STORAGE_CAPACITY][SOC_MAX][VALUE],
        initial_storage_level=dict_asset[STORAGE_CAPACITY][SOC_INITIAL][
            VALUE
        ],  # in terms of SOC #implication: balanced = True, ie. start=end
        inflow_conversion_factor=dict_asset[INPUT_POWER][EFFICIENCY][
            VALUE
        ],  # storing efficiency
        outflow_conversion_factor=dict_asset[OUTPUT_POWER][EFFICIENCY][
            VALUE
        ],  # efficiency of discharge
        invest_relation_input_capacity=dict_asset[INPUT_POWER][C_RATE][VALUE],
        # storage can be charged with invest_relation_output_capacity*capacity in one timeperiod
        invest_relation_output_capacity=dict_asset[OUTPUT_POWER][C_RATE][VALUE]
        # storage can be emptied with invest_relation_output_capacity*capacity in one timeperiod
    )
    model.add(storage)
    kwargs[OEMOF_GEN_STORAGE].update({dict_asset[LABEL]: storage})


def source_non_dispatchable_fix(model, dict_asset, **kwargs):
    r"""
    Defines a non dispatchable source with a fixed capacity.

    See :py:func:`~.source` for more information, including parameters.

    Notes
    -----
    Tested with:
    - test_source_non_dispatchable_fix()

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the source object.

    """
    outputs = {
        kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
            label=dict_asset[LABEL],
            fix=dict_asset[TIMESERIES],
            nominal_value=dict_asset[INSTALLED_CAP][VALUE],
            variable_costs=dict_asset[DISPATCH_PRICE][VALUE],
            emission_factor=dict_asset[EMISSION_FACTOR][VALUE],
        )
    }

    source_non_dispatchable = solph.Source(label=dict_asset[LABEL], outputs=outputs)

    model.add(source_non_dispatchable)
    kwargs[OEMOF_SOURCE].update({dict_asset[LABEL]: source_non_dispatchable})
    logging.debug(
        f"Added: Non-dispatchable source {dict_asset[LABEL]} (fixed capacity) to bus {dict_asset[OUTFLOW_DIRECTION]}.",
    )


def source_non_dispatchable_optimize(model, dict_asset, **kwargs):
    r"""
    Defines a non dispatchable source with a capacity to be optimized.

    See :py:func:`~.source` for more information, including parameters.

    Notes
    -----
    Tested with:
    - test_source_non_dispatchable_optimize()

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the source object.

    """
    if MAXIMUM_ADD_CAP_NORMALIZED in dict_asset:
        maximum = dict_asset[MAXIMUM_ADD_CAP_NORMALIZED][VALUE]
    else:
        maximum = dict_asset[MAXIMUM_ADD_CAP][VALUE]
    if INSTALLED_CAP_NORMALIZED in dict_asset:
        existing = dict_asset[INSTALLED_CAP_NORMALIZED][VALUE]
    else:
        existing = dict_asset[INSTALLED_CAP][VALUE]
    outputs = {
        kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
            label=dict_asset[LABEL],
            fix=dict_asset[TIMESERIES_NORMALIZED],
            investment=solph.Investment(
                ep_costs=dict_asset[SIMULATION_ANNUITY][VALUE]
                / dict_asset[TIMESERIES_PEAK][VALUE],
                maximum=maximum,
                existing=existing,
            ),
            # variable_costs are devided by time series peak as normalized time series are used as actual_value
            variable_costs=dict_asset[DISPATCH_PRICE][VALUE]
            / dict_asset[TIMESERIES_PEAK][VALUE],
            # add emission_factor for emission contraint
            emission_factor=dict_asset[EMISSION_FACTOR][VALUE],
        )
    }

    source_non_dispatchable = solph.Source(label=dict_asset[LABEL], outputs=outputs)

    model.add(source_non_dispatchable)
    kwargs[OEMOF_SOURCE].update({dict_asset[LABEL]: source_non_dispatchable})
    logging.debug(
        f"Added: Non-dispatchable source {dict_asset[LABEL]} (capacity to be optimized) to bus {dict_asset[OUTFLOW_DIRECTION]}."
    )


def source_dispatchable_optimize(model, dict_asset, **kwargs):
    r"""
    Defines a dispatchable source with a fixed capacity.

    See :py:func:`~.source` for more information, including parameters.

    Notes
    -----
    Tested with:
    - test_source_dispatchable_optimize_normalized_timeseries()
    - test_source_dispatchable_optimize_timeseries_not_normalized_timeseries()

     Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the source object.

    """
    if TIMESERIES_NORMALIZED in dict_asset:
        outputs = {
            kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
                label=dict_asset[LABEL],
                max=dict_asset[TIMESERIES_NORMALIZED],
                investment=solph.Investment(
                    ep_costs=dict_asset[SIMULATION_ANNUITY][VALUE]
                    / dict_asset[TIMESERIES_PEAK][VALUE],
                    maximum=dict_asset[MAXIMUM_ADD_CAP][VALUE],
                    existing=dict_asset[INSTALLED_CAP][VALUE],
                ),
                # variable_costs are devided by time series peak as normalized time series are used as actual_value
                variable_costs=dict_asset[DISPATCH_PRICE][VALUE]
                / dict_asset[TIMESERIES_PEAK][VALUE],
                # add emission_factor for emission contraint
                emission_factor=dict_asset[EMISSION_FACTOR][VALUE],
            )
        }
        source_dispatchable = solph.Source(label=dict_asset[LABEL], outputs=outputs,)
    else:
        if TIMESERIES in dict_asset:
            logging.info(
                f"Asset {dict_asset[LABEL]} is introduced as a dispatchable source with an availability schedule."
            )
            logging.debug(
                f"The availability schedule is solely introduced because the key {TIMESERIES_NORMALIZED} was not in the asset´s dictionary. \n"
                f"It should only be applied to DSO sources. "
                f"If the asset should not have this behaviour, please create an issue.",
            )
        outputs = {
            kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
                label=dict_asset[LABEL],
                investment=solph.Investment(
                    ep_costs=dict_asset[SIMULATION_ANNUITY][VALUE],
                    existing=dict_asset[INSTALLED_CAP][VALUE],
                    maximum=dict_asset[MAXIMUM_ADD_CAP][VALUE],
                ),
                variable_costs=dict_asset[DISPATCH_PRICE][VALUE],
                # add emission_factor for emission contraint
                emission_factor=dict_asset[EMISSION_FACTOR][VALUE],
            )
        }
        source_dispatchable = solph.Source(label=dict_asset[LABEL], outputs=outputs,)
    model.add(source_dispatchable)
    kwargs[OEMOF_SOURCE].update({dict_asset[LABEL]: source_dispatchable})
    logging.debug(
        f"Added: Dispatchable source {dict_asset[LABEL]} (capacity to be optimized) to bus {dict_asset[OUTFLOW_DIRECTION]}."
    )


def source_dispatchable_fix(model, dict_asset, **kwargs):
    r"""
    Defines a dispatchable source with a fixed capacity.

    See :py:func:`~.source` for more information, including parameters.

    Notes
    -----
    Tested with:
    - test_source_dispatchable_fix_normalized_timeseries()
    - test_source_dispatchable_fix_timeseries_not_normalized_timeseries()

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the source object.

    """
    if TIMESERIES_NORMALIZED in dict_asset:
        outputs = {
            kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
                label=dict_asset[LABEL],
                max=dict_asset[TIMESERIES_NORMALIZED],
                existing=dict_asset[INSTALLED_CAP][VALUE],
                variable_costs=dict_asset[DISPATCH_PRICE][VALUE],
                # add emission_factor for emission contraint
                emission_factor=dict_asset[EMISSION_FACTOR][VALUE],
            )
        }
        source_dispatchable = solph.Source(label=dict_asset[LABEL], outputs=outputs,)
    else:
        if TIMESERIES in dict_asset:
            logging.info(
                f"Asset {dict_asset[LABEL]} is introduced as a dispatchable source with an availability schedule."
            )
            logging.debug(
                f"The availability schedule is solely introduced because the key {TIMESERIES_NORMALIZED} was not in the asset´s dictionary. \n"
                f"It should only be applied to DSO sources. "
                f"If the asset should not have this behaviour, please create an issue.",
            )
        outputs = {
            kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION]]: solph.Flow(
                label=dict_asset[LABEL],
                existing=dict_asset[INSTALLED_CAP][VALUE],
                variable_costs=dict_asset[DISPATCH_PRICE][VALUE],
            )
        }
        source_dispatchable = solph.Source(label=dict_asset[LABEL], outputs=outputs,)
    model.add(source_dispatchable)
    kwargs[OEMOF_SOURCE].update({dict_asset[LABEL]: source_dispatchable})
    logging.debug(
        f"Added: Dispatchable source {dict_asset[LABEL]} (fixed capacity) to bus {dict_asset[OUTFLOW_DIRECTION]}."
    )


def sink_dispatchable_optimize(model, dict_asset, **kwargs):
    r"""
    Define a dispatchable sink.

    The dispatchable sink is capacity-optimized, without any costs connected to the capacity of the asset.
    Applications of this asset type are: Feed-in sink, excess sink.

    See :py:func:`~.sink` for more information, including parameters.

    Notes
    -----
    Tested with:
    - test_sink_dispatchable_single_input_bus()
    - test_sink_dispatchable_multiple_input_busses()

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the sink object.

    """
    # check if the sink has multiple input busses
    if isinstance(dict_asset[INFLOW_DIRECTION], list):
        inputs = {}
        index = 0
        for bus in dict_asset[INFLOW_DIRECTION]:
            inputs[kwargs[OEMOF_BUSSES][bus]] = solph.Flow(
                label=dict_asset[LABEL],
                variable_costs=dict_asset[DISPATCH_PRICE][VALUE][index],
                investment=solph.Investment(),
            )
            index += 1
    else:
        inputs = {
            kwargs[OEMOF_BUSSES][dict_asset[INFLOW_DIRECTION]]: solph.Flow(
                label=dict_asset[LABEL],
                variable_costs=dict_asset[DISPATCH_PRICE][VALUE],
                investment=solph.Investment(),
            )
        }

    # create and add excess electricity sink to micro_grid_system - variable
    sink_dispatchable = solph.Sink(label=dict_asset[LABEL], inputs=inputs,)
    model.add(sink_dispatchable)
    kwargs[OEMOF_SINK].update({dict_asset[LABEL]: sink_dispatchable})
    logging.debug(
        f"Added: Dispatchable sink {dict_asset[LABEL]} (to be capacity optimized) to bus {dict_asset[INFLOW_DIRECTION]}.",
    )


def sink_non_dispatchable(model, dict_asset, **kwargs):
    r"""
    Defines a non dispatchable sink.

    See :py:func:`~.sink` for more information, including parameters.

    Notes
    -----
    Tested with:
    - test_sink_non_dispatchable_single_input_bus()
    - test_sink_non_dispatchable_multiple_input_busses()

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the sink object.

    """
    # check if the sink has multiple input busses
    if isinstance(dict_asset[INFLOW_DIRECTION], list):
        inputs = {}
        index = 0
        for bus in dict_asset[INFLOW_DIRECTION]:
            inputs[kwargs[OEMOF_BUSSES][bus]] = solph.Flow(
                fix=dict_asset[TIMESERIES], nominal_value=1
            )
            index += 1
    else:
        inputs = {
            kwargs[OEMOF_BUSSES][dict_asset[INFLOW_DIRECTION]]: solph.Flow(
                fix=dict_asset[TIMESERIES], nominal_value=1
            )
        }

    # create and add demand sink to micro_grid_system - fixed
    sink_demand = solph.Sink(label=dict_asset[LABEL], inputs=inputs,)
    model.add(sink_demand)
    kwargs[OEMOF_SINK].update({dict_asset[LABEL]: sink_demand})
    logging.debug(
        f"Added: Non-dispatchable sink {dict_asset[LABEL]} to bus {dict_asset[INFLOW_DIRECTION]}"
    )


def chp_fix(model, dict_asset, **kwargs):
    r"""
    Extraction turbine chp from Oemof solph. Extraction turbine must have one input and two outputs
    Notes
    -----
    Tested with:
    - test_to_be_written()

    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the extraction turbine component.

    """

    inputs = {
        kwargs[OEMOF_BUSSES][dict_asset[INFLOW_DIRECTION]]: solph.Flow(
            nominal_value=dict_asset[INSTALLED_CAP][VALUE]
        )
    }

    busses_energy_vectors = [
        kwargs[OEMOF_BUSSES][b].energy_vector for b in dict_asset[OUTFLOW_DIRECTION]
    ]
    idx_el = busses_energy_vectors.index("Electricity")
    idx_th = busses_energy_vectors.index("Heat")
    el_bus = kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION][idx_el]]
    th_bus = kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION][idx_th]]

    outputs = {
        el_bus: solph.Flow(),
        th_bus: solph.Flow(),
    }  # if kW for heat and kW for elect then insert it under nominal_value

    beta = dict_asset[BETA][VALUE]

    efficiency_el_wo_heat_extraction = dict_asset[EFFICIENCY][VALUE][idx_th]
    efficiency_th_max_heat_extraction = dict_asset[EFFICIENCY][VALUE][idx_el]
    efficiency_el_max_heat_extraction = (
        efficiency_el_wo_heat_extraction - beta * efficiency_th_max_heat_extraction
    )
    efficiency_full_condensation = {el_bus: efficiency_el_wo_heat_extraction}

    efficiencies = {
        el_bus: efficiency_el_max_heat_extraction,
        th_bus: efficiency_th_max_heat_extraction,
    }

    ext_turb_chp = solph.components.ExtractionTurbineCHP(
        label=dict_asset[LABEL],
        inputs=inputs,
        outputs=outputs,
        conversion_factors=efficiencies,
        conversion_factor_full_condensation=efficiency_full_condensation,
    )

    model.add(ext_turb_chp)
    kwargs[OEMOF_ExtractionTurbineCHP].update({dict_asset[LABEL]: ext_turb_chp})


def chp_optimize(model, dict_asset, **kwargs):
    r"""
    Extraction turbine chp from Oemof solph. Extraction turbine must have one input and two outputs
    Notes
    -----
    Tested with:
    - test_to_be_written()
    
    Returns
    -------
    Indirectly updated `model` and dict of asset in `kwargs` with the extraction turbine component.

    """

    inputs = {kwargs[OEMOF_BUSSES][dict_asset[INFLOW_DIRECTION]]: solph.Flow()}

    busses_energy_vectors = [
        kwargs[OEMOF_BUSSES][b].energy_vector for b in dict_asset[OUTFLOW_DIRECTION]
    ]

    idx_el = busses_energy_vectors.index("Electricity")
    idx_th = busses_energy_vectors.index("Heat")
    el_bus = kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION][idx_el]]
    th_bus = kwargs[OEMOF_BUSSES][dict_asset[OUTFLOW_DIRECTION][idx_th]]

    outputs = {
        el_bus: solph.Flow(
            investment=solph.Investment(
                ep_costs=dict_asset[SIMULATION_ANNUITY][VALUE],
                maximum=dict_asset[MAXIMUM_ADD_CAP][VALUE],
                existing=dict_asset[INSTALLED_CAP][VALUE],
            )
        ),
        th_bus: solph.Flow(),
    }

    beta = dict_asset[BETA][VALUE]

    efficiency_el_wo_heat_extraction = dict_asset[EFFICIENCY][VALUE][idx_el]
    efficiency_th_max_heat_extraction = dict_asset[EFFICIENCY][VALUE][idx_th]
    efficiency_el_max_heat_extraction = (
        efficiency_el_wo_heat_extraction - beta * efficiency_th_max_heat_extraction
    )
    efficiency_full_condensation = {el_bus: efficiency_el_wo_heat_extraction}

    efficiencies = {
        el_bus: efficiency_el_max_heat_extraction,
        th_bus: efficiency_th_max_heat_extraction,
    }

    ext_turb_chp = solph.components.ExtractionTurbineCHP(
        label=dict_asset[LABEL],
        inputs=inputs,
        outputs=outputs,
        conversion_factors=efficiencies,
        conversion_factor_full_condensation=efficiency_full_condensation,
    )

    model.add(ext_turb_chp)
    kwargs[OEMOF_ExtractionTurbineCHP].update({dict_asset[LABEL]: ext_turb_chp})



"""
Module D2 - Model constraints
=============================

This module gathers all constraints that can be added to the MVS optimisation. 
we will probably require another input CSV file or further parameters in simulation_settings.

Future constraints are discussed in issue #133 (https://github.com/rl-institut/multi-vector-simulator/issues/133)

constraints should be tested in-code (examples) and by comparing the lp file generated.
"""
import logging
import pyomo.environ as po
from oemof.solph import constraints

from multi_vector_simulator.utils.constants import DEFAULT_WEIGHTS_ENERGY_CARRIERS

from multi_vector_simulator.utils.constants_json_strings import (
    OEMOF_SOURCE,
    OEMOF_SINK,
    OEMOF_BUSSES,
    ENERGY_PRODUCTION,
    ENERGY_PROVIDERS,
    ENERGY_CONSUMPTION,
    ENERGY_VECTOR,
    VALUE,
    LABEL,
    INFLOW_DIRECTION,
    OUTFLOW_DIRECTION,
    DSO_CONSUMPTION,
    RENEWABLE_SHARE_DSO,
    RENEWABLE_ASSET_BOOL,
    EXCESS_SINK,
    CONSTRAINTS,
    MINIMAL_RENEWABLE_FACTOR,
    MAXIMUM_EMISSIONS,
    MINIMAL_DEGREE_OF_AUTONOMY,
    DSO_FEEDIN,
    NET_ZERO_ENERGY,
)

# Keys for dicts renewable_assets and non_renewable_assets
OEMOF_SOLPH_OBJECT_ASSET = "object"
WEIGHTING_FACTOR_ENERGY_CARRIER = "weighting_factor_energy_carrier"
RENEWABLE_SHARE_ASSET_FLOW = "renewable_share_asset_flow"
OEMOF_SOLPH_OBJECT_BUS = "oemof_solph_object_bus"


def add_constraints(local_energy_system, dict_values, dict_model):
    r"""
    Adds all constraints activated in constraints.csv to the energy system model.

    Possible constraints:
    - Minimal renewable factor constraint
    Parameters
    ----------
    local_energy_system:  :oemof-solph: <oemof.solph.model>
        Energy system model generated from oemof-solph for the energy system scenario, including all energy system assets.

    dict_values: dict
        All simulation parameters

    dict_model: dict of :oemof-solph: <oemof.solph.assets>
        Dictionary including the oemof-solph component assets, which need to be connected with constraints

    Returns
    -------
    local_energy_system: :oemof-solph: <oemof.solph.model>
        Updated object local_energy_system with the additional constraints and bounds.

    Notes
    -----
    The constraints can be validated by evaluating the LP file.
    Additionally, there are validation tests in `E4_verification_of_constraints`.

    Tested with:
    - D2.test_add_constraints_maximum_emissions()
    - D2.test_add_constraints_maximum_emissions_None()
    - D2.test_add_constraints_minimal_renewable_share()
    - D2.test_add_constraints_minimal_renewable_share_is_0()
    - D2.test_add_constraints_net_zero_energy_requirement_is_true()
    - D2.test_add_constraints_net_zero_energy_requirement_is_false()
    """

    constraint_functions = {
        MINIMAL_RENEWABLE_FACTOR: constraint_minimal_renewable_share,
        MAXIMUM_EMISSIONS: constraint_maximum_emissions,
        MINIMAL_DEGREE_OF_AUTONOMY: constraint_minimal_degree_of_autonomy,
        NET_ZERO_ENERGY: constraint_net_zero_energy,
    }

    count_added_constraints = 0

    for constraint in dict_values[CONSTRAINTS]:
        # if the constraint is not within its proper range of admissible values, None is returned
        les = constraint_functions[constraint](
            local_energy_system, dict_values, dict_model
        )
        # if the contraint is not applied (because not defined by the user, or with a value which
        # is not in the acceptable range the constrain function will return None
        if les is not None:
            local_energy_system = les
            count_added_constraints += 1

    if count_added_constraints == 0:
        logging.info("No modelling constraint to be introduced.")
    else:
        logging.debug(f"Number of added constraints: {count_added_constraints}")

    return local_energy_system


def constraint_maximum_emissions(model, dict_values, dict_model=None):
    r"""
    Resulting in an energy system adhering to a maximum amount of emissions.

    Parameters
    ----------
    model: :oemof-solph: <oemof.solph.model>
        Model to which constraint is added.

    dict_values: dict
        All simulation parameters

    dict_model: None
        To match other constraint function's format

    Notes
    -----
    Tested with:
    - D2.test_constraint_maximum_emissions()

    """
    maximum_emissions = dict_values[CONSTRAINTS][MAXIMUM_EMISSIONS][VALUE]
    if maximum_emissions is not None:
        # Updates the model with the constraint for maximum amount of emissions
        constraints.emission_limit(model, limit=maximum_emissions)
        logging.info("Added maximum emission constraint.")
        answer = model
    else:
        answer = None

    return answer


def constraint_minimal_renewable_share(model, dict_values, dict_model):
    r"""
    Resulting in an energy system adhering to a minimal renewable factor.

    Please be aware that the renewable factor that has to adhere to the minimal renewable factor is not the one of one specific sector,
    but of the overall energy system. This means that eg. 1 kg H2 is produced renewably, it goes into account with a heavier
    weighting factor then one renewably produced electricity unit.

    Parameters
    ----------
    model: :oemof-solph: <oemof.solph.model>
        Model to which constraint is added.

    dict_values: dict
        All simulation parameters

    dict_model: dict of :oemof-solph: <oemof.solph.assets>
        Dictionary including the oemof-solph component assets, which need to be connected with constraints

    Notes
    -----
    The renewable factor of the whole energy system has to adhere for following constraint:

    .. math::
        minimal renewable factor <= \frac{\sum renewable generation \cdot weighting factor}{\sum renewable generation \cdot weighting factor + \sum non-renewable generation \cdot weighting factor}

    Tested with:
    - Test_Constraints.test_benchmark_minimal_renewable_share_constraint()
    """

    if dict_values[CONSTRAINTS][MINIMAL_RENEWABLE_FACTOR][VALUE] > 0:

        (
            renewable_assets,
            non_renewable_assets,
        ) = prepare_constraint_minimal_renewable_share(dict_values, dict_model,)

        def renewable_share_rule(model):
            renewable_generation = 0
            total_generation = 0

            # Get the flows from all renewable assets
            for asset in renewable_assets:
                generation = (
                    sum(
                        model.flow[
                            renewable_assets[asset][OEMOF_SOLPH_OBJECT_ASSET],
                            renewable_assets[asset][OEMOF_SOLPH_OBJECT_BUS],
                            :,
                        ]
                    )
                    * renewable_assets[asset][WEIGHTING_FACTOR_ENERGY_CARRIER]
                    * renewable_assets[asset][RENEWABLE_SHARE_ASSET_FLOW]
                )
                total_generation += generation
                renewable_generation += generation

            # Get the flows from all non renewable assets
            for asset in non_renewable_assets:
                generation = (
                    sum(
                        model.flow[
                            non_renewable_assets[asset][OEMOF_SOLPH_OBJECT_ASSET],
                            non_renewable_assets[asset][OEMOF_SOLPH_OBJECT_BUS],
                            :,
                        ]
                    )
                    * non_renewable_assets[asset][WEIGHTING_FACTOR_ENERGY_CARRIER]
                    * (1 - non_renewable_assets[asset][RENEWABLE_SHARE_ASSET_FLOW])
                )
                total_generation += generation

            expr = (
                renewable_generation
                - (dict_values[CONSTRAINTS][MINIMAL_RENEWABLE_FACTOR][VALUE])
                * total_generation
            )
            return expr >= 0

        model.constraint_minimal_renewable_share = po.Constraint(
            rule=renewable_share_rule
        )

        logging.info("Added minimal renewable factor constraint.")
        answer = model
    else:
        answer = None

    return answer


def prepare_constraint_minimal_renewable_share(
    dict_values, dict_model,
):
    r"""
    Prepare creating the minimal renewable factor constraint by processing dict_values

    Parameters
    ----------
    dict_values: dict
        All simulation parameters

    dict_model: dict of :oemof-solph: <oemof.solph.assets>
        Dictionary including the oemof-solph component assets, which need to be connected with constraints

    Returns
    -------
    renewable_assets: dict
        Dictionary of all assets with renewable generation in the energy system.
        Defined by: oemof_solph_object_asset, weighting_factor_energy_carrier, renewable_share_asset_flow, oemof_solph_object_bus

    non_renewable_assets: dict
        Dictionary of all assets with renewable generation in the energy system.
        Defined by: oemof_solph_object_asset, weighting_factor_energy_carrier, renewable_share_asset_flow, oemof_solph_object_bus

    """
    # dicts for processing
    renewable_assets = {}
    non_renewable_assets = {}

    # Determine which energy sources are of renewable origin and which are fossil-fuelled.
    # DSO sources are added separately (as they do not have parameter "RENEWABLE_ASSET_BOOL".
    assets_without_renewable_asset_bool = []
    for asset in dict_values[ENERGY_PRODUCTION]:
        if RENEWABLE_ASSET_BOOL in dict_values[ENERGY_PRODUCTION][asset]:
            if (
                dict_values[ENERGY_PRODUCTION][asset][RENEWABLE_ASSET_BOOL][VALUE]
                is True
            ):
                renewable_assets.update(
                    {
                        asset: {
                            OEMOF_SOLPH_OBJECT_ASSET: dict_model[OEMOF_SOURCE][
                                dict_values[ENERGY_PRODUCTION][asset][LABEL]
                            ],
                            OEMOF_SOLPH_OBJECT_BUS: dict_model[OEMOF_BUSSES][
                                dict_values[ENERGY_PRODUCTION][asset][OUTFLOW_DIRECTION]
                            ],
                            WEIGHTING_FACTOR_ENERGY_CARRIER: DEFAULT_WEIGHTS_ENERGY_CARRIERS[
                                dict_values[ENERGY_PRODUCTION][asset][ENERGY_VECTOR]
                            ][
                                VALUE
                            ],
                            RENEWABLE_SHARE_ASSET_FLOW: 1,
                        }
                    }
                )
            elif (
                dict_values[ENERGY_PRODUCTION][asset][RENEWABLE_ASSET_BOOL][VALUE]
                is False
            ):
                non_renewable_assets.update(
                    {
                        asset: {
                            OEMOF_SOLPH_OBJECT_ASSET: dict_model[OEMOF_SOURCE][
                                dict_values[ENERGY_PRODUCTION][asset][LABEL]
                            ],
                            OEMOF_SOLPH_OBJECT_BUS: dict_model[OEMOF_BUSSES][
                                dict_values[ENERGY_PRODUCTION][asset][OUTFLOW_DIRECTION]
                            ],
                            WEIGHTING_FACTOR_ENERGY_CARRIER: DEFAULT_WEIGHTS_ENERGY_CARRIERS[
                                dict_values[ENERGY_PRODUCTION][asset][ENERGY_VECTOR]
                            ][
                                VALUE
                            ],
                            RENEWABLE_SHARE_ASSET_FLOW: 0,
                        }
                    }
                )
            else:
                logging.warning(
                    f"Value of parameter {RENEWABLE_ASSET_BOOL} of asset {asset} is {dict_values[ENERGY_PRODUCTION][asset][RENEWABLE_ASSET_BOOL][VALUE]}, but should be True/False."
                )
        else:
            assets_without_renewable_asset_bool.append(asset)

    # This message is printed so that errors can be identified easier.
    assets_without_renewable_asset_bool_string = ", ".join(
        map(str, assets_without_renewable_asset_bool)
    )
    logging.debug(
        f"Following assets do not have key RENEWABLE_ASSET_BOOL: {assets_without_renewable_asset_bool_string}. "
        f"They should be all DSO consumption sources."
    )

    dso_sources = []
    for dso in dict_values[ENERGY_PROVIDERS]:
        # Get source connected to the specific DSO in question
        DSO_source_name = dso + DSO_CONSUMPTION
        # Add DSO to both renewable and nonrenewable assets (as only a share of their supply may be renewable)
        dso_sources.append(DSO_source_name)

        renewable_assets.update(
            {
                DSO_source_name: {
                    OEMOF_SOLPH_OBJECT_ASSET: dict_model[OEMOF_SOURCE][
                        dict_values[ENERGY_PRODUCTION][DSO_source_name][LABEL]
                    ],
                    OEMOF_SOLPH_OBJECT_BUS: dict_model[OEMOF_BUSSES][
                        dict_values[ENERGY_PRODUCTION][DSO_source_name][
                            OUTFLOW_DIRECTION
                        ]
                    ],
                    WEIGHTING_FACTOR_ENERGY_CARRIER: DEFAULT_WEIGHTS_ENERGY_CARRIERS[
                        dict_values[ENERGY_PRODUCTION][DSO_source_name][ENERGY_VECTOR]
                    ][VALUE],
                    RENEWABLE_SHARE_ASSET_FLOW: dict_values[ENERGY_PROVIDERS][dso][
                        RENEWABLE_SHARE_DSO
                    ][VALUE],
                }
            }
        )
        non_renewable_assets.update(
            {
                DSO_source_name: {
                    OEMOF_SOLPH_OBJECT_ASSET: dict_model[OEMOF_SOURCE][
                        dict_values[ENERGY_PRODUCTION][DSO_source_name][LABEL]
                    ],
                    OEMOF_SOLPH_OBJECT_BUS: dict_model[OEMOF_BUSSES][
                        dict_values[ENERGY_PRODUCTION][DSO_source_name][
                            OUTFLOW_DIRECTION
                        ]
                    ],
                    WEIGHTING_FACTOR_ENERGY_CARRIER: DEFAULT_WEIGHTS_ENERGY_CARRIERS[
                        dict_values[ENERGY_PRODUCTION][DSO_source_name][ENERGY_VECTOR]
                    ][VALUE],
                    RENEWABLE_SHARE_ASSET_FLOW: dict_values[ENERGY_PROVIDERS][dso][
                        RENEWABLE_SHARE_DSO
                    ][VALUE],
                }
            }
        )

    for k in assets_without_renewable_asset_bool:
        if k not in dso_sources:
            logging.warning(
                "There is an asset {k} that does not have any information whether it is of renewable origin or not. "
                "It is neither a DSO source nor a renewable or fuel source."
            )

    renewable_asset_string = ", ".join(map(str, renewable_assets.keys()))
    non_renewable_asset_string = ", ".join(map(str, non_renewable_assets.keys()))

    logging.debug(f"Data considered for the minimal renewable factor constraint:")
    logging.debug(f"Assets connected to renewable generation: {renewable_asset_string}")
    logging.debug(
        f"Assets connected to non-renewable generation: {non_renewable_asset_string}"
    )
    return renewable_assets, non_renewable_assets


def constraint_minimal_degree_of_autonomy(model, dict_values, dict_model):
    r"""
    Resulting in an energy system adhering to a minimal degree of autonomy.

    Please be aware that the minimal degree of autonomy is not applied to each sector individually,
    but to the overall energy system (via energy carrier weighting).

    Parameters
    ----------
    model: :oemof-solph: <oemof.solph.model>
        Model to which constraint is added.

    dict_values: dict
        All simulation parameters

    dict_model: dict of :oemof-solph: <oemof.solph.assets>
        Dictionary including the oemof-solph component assets, which need to be connected with constraints

    Notes
    -----
    The renewable factor of the whole energy system has to adhere for following constraint:

    .. math::
        minimal degree of autonomy \cdot (\sum local demand  \cdot weighting factor) <= \sum local demand \cdot weighting factor - \sum consumtion from energy providers \cdot weighting factor

    Tested with:
    - Test_Constraints.test_benchmark_minimal_degree_of_autonomy()
    """

    logging.debug(f"Data considered for the minimal degree of autonomy constraint:")

    if dict_values[CONSTRAINTS][MINIMAL_DEGREE_OF_AUTONOMY][VALUE] > 0:

        demands = prepare_demand_assets(dict_values, dict_model,)

        energy_provider_consumption_sources = prepare_energy_provider_consumption_sources(
            dict_values, dict_model,
        )

        def degree_of_autonomy_rule(model):
            total_demand = 0
            total_consumption_from_energy_provider = 0

            # Get the flows from demands and add weighing
            for asset in demands:
                demand_one_asset = (
                    sum(
                        model.flow[
                            demands[asset][OEMOF_SOLPH_OBJECT_BUS],
                            demands[asset][OEMOF_SOLPH_OBJECT_ASSET],
                            :,
                        ]
                    )
                    * demands[asset][WEIGHTING_FACTOR_ENERGY_CARRIER]
                )
                total_demand += demand_one_asset

            # Get the flows from providers and add weighing
            for asset in energy_provider_consumption_sources:
                consumption_of_one_provider = (
                    sum(
                        model.flow[
                            energy_provider_consumption_sources[asset][
                                OEMOF_SOLPH_OBJECT_ASSET
                            ],
                            energy_provider_consumption_sources[asset][
                                OEMOF_SOLPH_OBJECT_BUS
                            ],
                            :,
                        ]
                    )
                    * energy_provider_consumption_sources[asset][
                        WEIGHTING_FACTOR_ENERGY_CARRIER
                    ]
                )
                total_consumption_from_energy_provider += consumption_of_one_provider

            expr = (
                1 - dict_values[CONSTRAINTS][MINIMAL_DEGREE_OF_AUTONOMY][VALUE]
            ) * total_demand - total_consumption_from_energy_provider
            return expr >= 0

        model.constraint_minimal_degree_of_autonomy = po.Constraint(
            rule=degree_of_autonomy_rule
        )

        logging.info("Added minimal degree of autonomy constraint.")
        answer = model
    else:
        answer = None

    return answer


def prepare_demand_assets(
    dict_values, dict_model,
):
    r"""
    Perpare demand assets by processing `dict_values`

    Used for the following constraints:
    - minimal degree of autonomy

    Parameters
    ----------
    dict_values: dict
        All simulation parameters

    dict_model: dict of :oemof-solph: <oemof.solph.assets>
        Dictionary including the oemof-solph component assets, which need to be connected with constraints

    Notes
    -----
    Tested with:
    - test_prepare_demand_assets()

    Returns
    -------
    demands: dict
        Dictionary of all assets with all demands in the energy system.
        Defined by: oemof_solph_object_asset, weighting_factor_energy_carrier, oemof_solph_object_bus

    """
    # dicts for processing
    demands = {}
    demand_list = []

    # Determine energy demands
    for asset in dict_values[ENERGY_CONSUMPTION]:
        # Do not add flows into excess sink of feedin sink to the demands to be supplied
        if EXCESS_SINK not in asset and DSO_FEEDIN not in asset:
            demands.update(
                {
                    asset: {
                        OEMOF_SOLPH_OBJECT_ASSET: dict_model[OEMOF_SINK][
                            dict_values[ENERGY_CONSUMPTION][asset][LABEL]
                        ],
                        OEMOF_SOLPH_OBJECT_BUS: dict_model[OEMOF_BUSSES][
                            dict_values[ENERGY_CONSUMPTION][asset][INFLOW_DIRECTION]
                        ],
                        WEIGHTING_FACTOR_ENERGY_CARRIER: DEFAULT_WEIGHTS_ENERGY_CARRIERS[
                            dict_values[ENERGY_CONSUMPTION][asset][ENERGY_VECTOR]
                        ][
                            VALUE
                        ],
                    }
                }
            )
            demand_list.append(asset)

    # Preprocessing for log messages
    demand_string = ", ".join(map(str, demand_list))

    logging.debug(f"Demands: {demand_string}")
    return demands


def prepare_energy_provider_consumption_sources(
    dict_values, dict_model,
):
    r"""
    Prepare energy provider consumption sources by processing `dict_values`.

    Used for the following constraints:
    - minimal degree of autonomy
    - net zero energy (NZE)

    Parameters
    ----------
    dict_values: dict
        All simulation parameters

    dict_model: dict of :oemof-solph: <oemof.solph.assets>
        Dictionary including the oemof-solph component assets, which need to be connected with constraints

    Notes
    -----
    Tested with:
    - test_prepare_energy_provider_consumption_sources()

    Returns
    -------
    energy_provider_consumption_sources: dict
        Dictionary of all assets that are sources for the energy consumption from energy providers in the energy system.
        Defined by: oemof_solph_object_asset, weighting_factor_energy_carrier, oemof_solph_object_bus

    """
    # dicts for processing
    energy_provider_consumption_sources = {}
    dso_consumption_source_list = []

    for dso in dict_values[ENERGY_PROVIDERS]:
        # Get source connected to the specific DSO in question
        DSO_source_name = dso + DSO_CONSUMPTION
        # Add DSO to assets
        dso_consumption_source_list.append(DSO_source_name)

        energy_provider_consumption_sources.update(
            {
                DSO_source_name: {
                    OEMOF_SOLPH_OBJECT_ASSET: dict_model[OEMOF_SOURCE][
                        dict_values[ENERGY_PRODUCTION][DSO_source_name][LABEL]
                    ],
                    OEMOF_SOLPH_OBJECT_BUS: dict_model[OEMOF_BUSSES][
                        dict_values[ENERGY_PRODUCTION][DSO_source_name][
                            OUTFLOW_DIRECTION
                        ]
                    ],
                    WEIGHTING_FACTOR_ENERGY_CARRIER: DEFAULT_WEIGHTS_ENERGY_CARRIERS[
                        dict_values[ENERGY_PRODUCTION][DSO_source_name][ENERGY_VECTOR]
                    ][VALUE],
                }
            }
        )

    # Preprocessing for log messages
    dso_consumption_source_string = ", ".join(map(str, dso_consumption_source_list))

    logging.debug(
        f"Energy provider consumption sources: {dso_consumption_source_string}"
    )
    return energy_provider_consumption_sources


def prepare_energy_provider_feedin_sinks(
    dict_values, dict_model,
):
    r"""
    Prepare energy provider feedin sinks by processing `dict_values`.

    Used for the following constraints:
    - net zero energy (NZE)

    Parameters
    ----------
    dict_values: dict
        All simulation parameters

    dict_model: dict of :oemof-solph: <oemof.solph.assets>
        Dictionary including the oemof-solph component assets, which need to be connected with constraints

    Notes
    -----
    Tested with:
        - test_prepare_energy_provider_feedin_sinks()

    Returns
    -------
    energy_provider_feedin_sinks: dict
        Dictionary of all assets that are sinks for the energy feed-in to energy providers in the energy system.
        Defined by: oemof_solph_object_asset, weighting_factor_energy_carrier, oemof_solph_object_bus

    """
    # dicts for processing
    energy_provider_feedin_sinks = {}
    dso_feedin_sink_list = []

    for dso in dict_values[ENERGY_PROVIDERS]:
        # Get sink connected to the specific DSO in question
        DSO_sink_name = dso + DSO_FEEDIN
        # Add DSO to assets
        dso_feedin_sink_list.append(DSO_sink_name)

        energy_provider_feedin_sinks.update(
            {
                DSO_sink_name: {
                    OEMOF_SOLPH_OBJECT_ASSET: dict_model[OEMOF_SINK][
                        dict_values[ENERGY_CONSUMPTION][DSO_sink_name][LABEL]
                    ],
                    OEMOF_SOLPH_OBJECT_BUS: dict_model[OEMOF_BUSSES][
                        dict_values[ENERGY_CONSUMPTION][DSO_sink_name][INFLOW_DIRECTION]
                    ],
                    WEIGHTING_FACTOR_ENERGY_CARRIER: DEFAULT_WEIGHTS_ENERGY_CARRIERS[
                        dict_values[ENERGY_CONSUMPTION][DSO_sink_name][ENERGY_VECTOR]
                    ][VALUE],
                }
            }
        )

    # Preprocessing for log messages
    dso_feedin_sink_string = ", ".join(map(str, dso_feedin_sink_list))

    logging.debug(f"Energy provider feedin sinks: {dso_feedin_sink_string}")
    return energy_provider_feedin_sinks


def constraint_net_zero_energy(model, dict_values, dict_model):
    r"""
    Resulting in an energy system that is a net zero energy (NZE) or plus energy system.

    Please be aware that the NZE constraint is not applied to each sector individually,
    but to the overall energy system (via energy carrier weighting).

    Parameters
    ----------
    model: :oemof-solph: <oemof.solph.model>
        Model to which constraint is added.

    dict_values: dict
        All simulation parameters

    dict_model: dict of :oemof-solph: <oemof.solph.assets>
        Dictionary including the oemof-solph component assets, which need to be connected with constraints

    Notes
    -----
    The constraint reads as follows:

    .. math::
        \sum_{i} {E_{feedin, DSO} (i) \cdot w_i - E_{consumption, DSO} (i) \cdot w_i} >= 0

    Tested with:
    - Test_Constraints.test_net_zero_energy_constraint()
    """

    logging.debug(f"Data considered for the net zero energy (NZE) constraint:")

    if dict_values[CONSTRAINTS][NET_ZERO_ENERGY][VALUE] == True:

        energy_provider_feedin_sinks = prepare_energy_provider_feedin_sinks(
            dict_values, dict_model,
        )

        energy_provider_consumption_sources = prepare_energy_provider_consumption_sources(
            dict_values, dict_model,
        )

        def net_zero_energy(model):
            total_feedin_to_energy_provider = 0
            total_consumption_from_energy_provider = 0

            # Get the flows from provider sources and add weighing
            for asset in energy_provider_consumption_sources:
                consumption_of_one_provider = (
                    sum(
                        model.flow[
                            energy_provider_consumption_sources[asset][
                                OEMOF_SOLPH_OBJECT_ASSET
                            ],
                            energy_provider_consumption_sources[asset][
                                OEMOF_SOLPH_OBJECT_BUS
                            ],
                            :,
                        ]
                    )
                    * energy_provider_consumption_sources[asset][
                        WEIGHTING_FACTOR_ENERGY_CARRIER
                    ]
                )
                total_consumption_from_energy_provider += consumption_of_one_provider

            # Get the flows from provider sources and add weighing
            for asset in energy_provider_feedin_sinks:
                feedin_of_one_provider = (
                    sum(
                        model.flow[
                            energy_provider_feedin_sinks[asset][OEMOF_SOLPH_OBJECT_BUS],
                            energy_provider_feedin_sinks[asset][
                                OEMOF_SOLPH_OBJECT_ASSET
                            ],
                            :,
                        ]
                    )
                    * energy_provider_feedin_sinks[asset][
                        WEIGHTING_FACTOR_ENERGY_CARRIER
                    ]
                )
                total_feedin_to_energy_provider += feedin_of_one_provider

            expr = (
                total_feedin_to_energy_provider - total_consumption_from_energy_provider
            )
            return expr >= 0

        model.constraint_net_zero_energy = po.Constraint(rule=net_zero_energy)

        logging.info("Added net zero energy (NZE) constraint.")
        answer = model
    else:
        answer = None

    return answer



r"""
Module F1 - Plotting
====================

Module F1 describes all the functions that create plots.

- creating graphs for energy flows
- creating bar chart for capacity
- creating pie chart for cost data
- creating network graph for the model brackets only working on Ubuntu
"""

import logging
import os
import textwrap
import numpy as np

import pandas as pd

PLOTLY_INSTALLED = False
try:
    import plotly.graph_objs as go
    import plotly.express as px

    PLOTLY_INSTALLED = True
except ModuleNotFoundError:
    logging.warning(
        "You have installed the minimal configuration, if you want to output images "
        "please run the command <TODO>"
    )

import graphviz
import oemof
import oemof.solph as solph

from multi_vector_simulator.utils.constants import (
    PROJECT_DATA,
    ECONOMIC_DATA,
    LABEL,
    OUTPUT_FOLDER,
    PATHS_TO_PLOTS,
    PLOT_SANKEY,
    SOC,
)

from multi_vector_simulator.utils.constants_json_strings import (
    PROJECT_NAME,
    SCENARIO_NAME,
    CURR,
    KPI,
    UNIT,
    ENERGY_CONSUMPTION,
    TIMESERIES,
    DISPATCHABILITY,
    ENERGY_PRODUCTION,
    SIMULATION_SETTINGS,
    OPTIMIZED_ADD_CAP,
    TOTAL_FLOW,
    ANNUAL_TOTAL_FLOW,
    PEAK_FLOW,
    AVERAGE_FLOW,
    KPI_SCALAR_MATRIX,
    OPTIMIZED_FLOWS,
    DEMANDS,
    RESOURCES,
    TIME_INDEX,
)

from multi_vector_simulator.E1_process_results import (
    convert_demand_to_dataframe,
    convert_costs_to_dataframe,
)


def convert_plot_data_to_dataframe(plot_data_dict, data_type):
    """

    Parameters
    ----------
    plot_data_dict: dict
        timeseries for either demand or supply

    data_type: str
        one of DEMANDS or RESOURCES

    Returns
    -------
    df: pandas:`pandas.DataFrame<frame>`,
        timeseries for plotting
    """
    # Later, this dataframe can be passed to a function directly make the graphs with Plotly
    df = pd.DataFrame.from_dict(plot_data_dict[data_type], orient="columns")

    # Change the index of the dataframe
    df.reset_index(level=0, inplace=True)
    # Rename the timestamp column from 'index' to 'timestamp'
    df = df.rename(columns={"index": "timestamp"})
    return df


def extract_plot_data_and_title(dict_values, df_dem=None):
    """Dataframe used for the plots of demands and resources timeseries in the report

    Parameters
    ----------
    dict_values: dict
        output values of MVS

    df_dem: :class:`pandas.DataFrame<frame>`
        summarized demand information for each demand

    Returns
    -------
    :class:`pandas.DataFrame<frame>`

    """
    if df_dem is None:
        df_dem = convert_demand_to_dataframe(dict_values)

    # Collect the keys of various resources (PV, Wind, etc.)
    resources = dict_values[ENERGY_PRODUCTION].copy()
    res_keys = [k for k in resources.keys() if resources[k][DISPATCHABILITY] is False]

    # Gather all the keys of the various plots for later use in the graphOptions.csv
    dict_for_plots = {DEMANDS: {}, RESOURCES: {}}
    dict_plot_labels = {}

    # Add all the demands to the dict_for_plots dictionary, including the timeseries values
    for demand in df_dem.Demands:
        dict_for_plots[DEMANDS].update(
            {demand: dict_values[ENERGY_CONSUMPTION][demand][TIMESERIES]}
        )
        dict_plot_labels.update(
            {demand: dict_values[ENERGY_CONSUMPTION][demand][LABEL]}
        )

    # Add all the resources to the dict_for_plots dictionary, including the timeseries values
    for resource in res_keys:
        dict_for_plots[RESOURCES].update(
            {resource: dict_values[ENERGY_PRODUCTION][resource][TIMESERIES]}
        )
        dict_plot_labels.update(
            {resource: dict_values[ENERGY_PRODUCTION][resource][LABEL]}
        )

    return dict_for_plots, dict_plot_labels


def fixed_width_text(text, char_num=10):
    """Add linebreaks every char_num characters in a given text.

    Parameters
    ----------
    text: obj:'str'
        text to apply the linebreaks
    char_num: obj:'int'
        max number of characters in a line before a line break
        Default: 10
    Returns
    -------
    obj:'str'
        the text with line breaks after every char_num characters

    """
    # total number of characters in the text
    text_length = len(text)
    # integer number of lines of `char_num` character
    n_lines = int(text_length / char_num)
    # number of character in the last line
    last_line_length = text_length % char_num

    # split the text in lines of `char_num` character
    split_text = []
    for i in range(n_lines):
        split_text.append(text[(i * char_num) : ((i + 1) * char_num)])

    # I if the last line is not empty
    if n_lines > 0:
        if last_line_length > 0:
            split_text.append(text[((i + 1) * char_num) :])
        answer = "\n".join(split_text)
    else:
        answer = text
    return answer


class ESGraphRenderer:
    def __init__(
        self,
        energy_system=None,
        filepath="network",
        img_format=None,
        legend=True,
        txt_width=10,
        txt_fontsize=10,
        **kwargs,
    ):
        """Draw the energy system with Graphviz.

        Parameters
        ----------
        energy_system: `oemof.solph.network.EnergySystem`
            The oemof energy stystem

        filepath: str
            path, where the rendered result shall be saved, if an extension is provided, the format
            will be automatically adapted except if the `img_format` argument is provided
            Default: "network"

        img_format: str
            extension of the available image formats of graphviz (e.g "png", "svg", "pdf", ... )
            Default: "pdf"

        legend: bool
            specify, whether a legend will be added to the graph or not
            Default: False

        txt_width: int
            max number of characters in a line before a line break
            Default: 10

         txt_fontsize: int
            fontsize of the image's text (components labels)
            Default: 10

        Returns
        -------
        None:
        render the generated dot graph in the filepath

        Notes
        -----
        When new oemof-solph asset types are added to the available types in the MVS, this function needs to be updated, so that it can render the asset in the graph.
        """
        file_name, file_ext = os.path.splitext(filepath)

        self.energy_system = energy_system

        if img_format is None:
            if file_ext != "":
                img_format = file_ext.replace(".", "")
            else:
                img_format = "pdf"

        self.dot = graphviz.Digraph(filename=file_name, format=img_format, **kwargs)
        self.txt_width = txt_width
        self.txt_fontsize = str(txt_fontsize)
        self.busses = []

        if legend is True:
            with self.dot.subgraph(name="cluster_1") as c:
                # color of the legend box
                c.attr(color="black")
                # title of the legend box
                c.attr(label="Legends")
                self.add_bus(subgraph=c)
                self.add_sink(subgraph=c)
                self.add_source(subgraph=c)
                self.add_transformer(subgraph=c)
                self.add_storage(subgraph=c)

        # draw a node for each of the network's component. The shape depends on the component's type
        for nd in self.energy_system.nodes:
            if isinstance(nd, oemof.solph.network.Bus):
                self.add_bus(nd.label)
                # keep the bus reference for drawing edges later
                self.busses.append(nd)
            elif isinstance(nd, oemof.solph.network.Sink):
                self.add_sink(nd.label)
            elif isinstance(nd, oemof.solph.network.Source):
                self.add_source(nd.label)
            elif isinstance(nd, oemof.solph.network.Transformer):
                self.add_transformer(nd.label)
            elif isinstance(nd, oemof.solph.components.GenericStorage):
                self.add_storage(nd.label)
            else:
                logging.warning(
                    "The component {} of type {} is not implemented in the rendering "
                    "function of the energy model network graph drawer. It will be "
                    "rendered as an ellipse".format(nd.label, type(nd))
                )
                self.add_component(nd.label)

        # draw the edges between the nodes based on each bus inputs/outputs
        for bus in self.busses:
            for component in bus.inputs:
                # draw an arrow from the component to the bus
                self.connect(component, bus)
            for component in bus.outputs:
                # draw an arrow from the bus to the component
                self.connect(bus, component)

    def add_bus(self, label="Bus", subgraph=None):
        if subgraph is None:
            dot = self.dot
        else:
            dot = subgraph
        dot.node(
            label,
            shape="rectangle",
            fontsize="10",
            fixedsize="shape",
            width="4.1",
            height="0.3",
            style="filled",
            color="lightgrey",
        )

    def add_sink(self, label="Sink", subgraph=None):
        if subgraph is None:
            dot = self.dot
        else:
            dot = subgraph
        dot.node(
            fixed_width_text(label, char_num=self.txt_width),
            shape="trapezium",
            fontsize=self.txt_fontsize,
        )

    def add_source(self, label="Source", subgraph=None):
        if subgraph is None:
            dot = self.dot
        else:
            dot = subgraph
        dot.node(
            fixed_width_text(label, char_num=self.txt_width),
            shape="invtrapezium",
            fontsize=self.txt_fontsize,
        )

    def add_transformer(self, label="Transformer", subgraph=None):
        if subgraph is None:
            dot = self.dot
        else:
            dot = subgraph
        dot.node(
            fixed_width_text(label, char_num=self.txt_width),
            shape="rectangle",
            fontsize=self.txt_fontsize,
        )

    def add_storage(self, label="Storage", subgraph=None):
        if subgraph is None:
            dot = self.dot
        else:
            dot = subgraph
        dot.node(
            fixed_width_text(label, char_num=self.txt_width),
            shape="rectangle",
            style="rounded",
            fontsize=self.txt_fontsize,
        )

    def add_component(self, label="component", subgraph=None):
        if subgraph is None:
            dot = self.dot
        else:
            dot = subgraph
        dot.node(
            fixed_width_text(label, char_num=self.txt_width),
            fontsize=self.txt_fontsize,
        )

    def connect(self, a, b):
        """Draw an arrow from node a to node b

        Parameters
        ----------
        a: `oemof.solph.network.Node`
            An oemof node (usually a Bus or a Component)

        b: `oemof.solph.network.Node`
            An oemof node (usually a Bus or a Component)
        """
        if not isinstance(a, oemof.solph.network.Bus):
            a = fixed_width_text(a.label, char_num=self.txt_width)
        else:
            a = a.label
        if not isinstance(b, oemof.solph.network.Bus):
            b = fixed_width_text(b.label, char_num=self.txt_width)
        else:
            b = b.label

        self.dot.edge(a, b)

    def view(self, **kwargs):
        """Call the view method of the DiGraph instance"""
        self.dot.view(**kwargs)

    def render(self, **kwargs):
        """Call the render method of the DiGraph instance"""
        self.dot.render(**kwargs)

    def sankey(self, results):
        """Return a dict to a plotly sankey diagram"""
        busses = []

        labels = []
        sources = []
        targets = []
        values = []

        # bus_data.update({bus: solph.views.node(results_main, bus)})

        # draw a node for each of the network's component. The shape depends on the component's type
        for nd in self.energy_system.nodes:
            if isinstance(nd, oemof.solph.network.Bus):

                # keep the bus reference for drawing edges later
                bus = nd
                busses.append(bus)

                bus_label = bus.label

                labels.append(nd.label)

                flows = solph.views.node(results, bus_label)["sequences"]

                # draw an arrow from the component to the bus
                for component in bus.inputs:
                    if component.label not in labels:
                        labels.append(component.label)

                    sources.append(labels.index(component.label))
                    targets.append(labels.index(bus_label))

                    val = flows[((component.label, bus_label), "flow")].sum()
                    # if val == 0:
                    #     val = 1
                    values.append(val)

                for component in bus.outputs:
                    # draw an arrow from the bus to the component
                    if component.label not in labels:
                        labels.append(component.label)

                    sources.append(labels.index(bus_label))
                    targets.append(labels.index(component.label))

                    val = flows[((bus_label, component.label), "flow")].sum()

                    # if val == 0:
                    #     val = 1
                    values.append(val)

        fig = go.Figure(
            data=[
                go.Sankey(
                    node=dict(
                        pad=15,
                        thickness=20,
                        line=dict(color="black", width=0.5),
                        label=labels,
                        hovertemplate="Node has total value %{value}<extra></extra>",
                        color="blue",
                    ),
                    link=dict(
                        source=sources,  # indices correspond to labels, eg A1, A2, A2, B1, ...
                        target=targets,
                        value=values,
                        hovertemplate="Link from node %{source.label}<br />"
                        + "to node%{target.label}<br />has value %{value}"
                        + "<br />and data <extra></extra>",
                    ),
                )
            ]
        )

        fig.update_layout(title_text="Basic Sankey Diagram", font_size=10)
        return fig.to_dict()


def get_color(idx_line, color_list=None):
    """Pick a color within a color list with periodic boundary conditions

    Parameters
    ----------
    idx_line: int
        index of the line in a plot for which a color is required

    colors: list of str or list to tuple (hexadecimal or rbg code)
        list of colors
        Default: None

    Returns
    -------
    The color in the color list corresponding to the index modulo the color list length

    """
    if color_list is None:
        color_list = (
            "#1f77b4",
            "#ff7f0e",
            "#2ca02c",
            "#d62728",
            "#9467bd",
            "#8c564b",
            "#e377c2",
            "#7f7f7f",
            "#bcbd22",
            "#17becf",
        )
    n_colors = len(color_list)
    return color_list[idx_line % n_colors]


def save_plots_to_disk(
    fig_obj, file_name, file_path="", width=None, height=None, scale=None
):
    r"""
    This function saves the plots generated using the Plotly library in this module to the outputs folder.

    Parameters
    ----------
    fig_obj: instance of the classes of the Plotly go library used to generate the plots in this auto-report
        Figure object of the plotly plots

    file_name: str
        The name of the PNG image of the plot to be saved in the output folder.

    file_path: str
        Path where the image shall be saved

    width: int or float
        The width of the picture to be saved in pixels.
        Default: None

    height: int or float
        The height of the picture to be saved in pixels.
        Default: None

    scale: int or float
        The scale by which the plotly image ought to be multiplied.
        Default: None

    Returns
    -------
    Nothing is returned. This function call results in the plots being saved as .png images to the disk.
    """

    if not file_name.endswith("png"):
        file_name = file_name + ".png"

    logging.info("Saving {} under {}".format(file_name, file_path))

    file_path_out = os.path.join(file_path, file_name)
    with open(file_path_out, "wb") as fp:
        fig_obj.write_image(fp, width=width, height=height, scale=scale)


def get_fig_style_dict():
    styling_dict = dict(
        showgrid=True,
        gridwidth=1.5,
        zeroline=True,
        autorange=True,
        linewidth=1,
        ticks="inside",
        title_font=dict(size=18, color="black"),
    )
    return styling_dict


def create_plotly_line_fig(
    x_data,
    y_data,
    plot_title=None,
    x_axis_name=None,
    y_axis_name=None,
    color_for_plot="#0A2342",
    file_path=None,
):
    r"""
    Create figure for generic timeseries lineplots

    Parameters
    ----------
    x_data: list, or pandas series
        The list of abscissas of the data required for plotting.

    y_data: list, or pandas series, or list of lists
        The list of ordinates of the data required for plotting.

    plot_title: str
        The title of the plot generated.
        Default: None

    x_axis_name: str
        Default: None

    y_axis_name: str
        Default: None

    file_path: str
        Path where the image shall be saved if not None

    Returns
    -------
    fig :class:`plotly.graph_objs.Figure`
        figure object
    """
    fig = go.Figure()

    styling_dict = get_fig_style_dict()
    styling_dict["mirror"] = True

    fig.add_trace(
        go.Scatter(
            x=x_data,
            y=y_data,
            mode="lines",
            line=dict(color=color_for_plot, width=2.5),
        )
    )
    fig.update_layout(
        xaxis_title=x_axis_name,
        yaxis_title=y_axis_name,
        template="simple_white",
        xaxis=styling_dict,
        yaxis=styling_dict,
        font_family="sans-serif",
        title={
            "text": plot_title,
            "y": 0.90,
            "x": 0.5,
            "font_size": 23,
            "xanchor": "center",
            "yanchor": "top",
        },
    )

    name_file = "input_timeseries_" + plot_title + ".png"

    if file_path is not None:

        # Function call to save the Plotly plot to the disk
        save_plots_to_disk(
            fig_obj=fig,
            file_path=file_path,
            file_name=name_file,
            width=1200,
            height=600,
            scale=5,
        )

    return fig


def plot_timeseries(
    dict_values,
    data_type=DEMANDS,
    sector_demands=None,
    max_days=None,
    color_list=None,
    file_path=None,
):
    r"""Plot timeseries as line chart.

    Parameters
    ----------
    dict_values :
        dict Of all input and output parameters up to F0

    data_type: str
        one of DEMANDS or RESOURCES
        Default: DEMANDS

    sector_demands: str
        Name of the sector of the energy system
        Default: None

    max_days: int
        maximal number of days the timeseries should be displayed for

    color_list: list of str or list to tuple (hexadecimal or rbg code)
        list of colors
        Default: None

    file_path: str
        Path where the image shall be saved if not None
        Default: None

    Returns
    -------
    Dict with html DOM id for the figure as key and :class:`plotly.graph_objs.Figure` as value
    """

    df_dem = convert_demand_to_dataframe(
        dict_values=dict_values, sector_demands=sector_demands
    )
    dict_for_plots, dict_plot_labels = extract_plot_data_and_title(
        dict_values, df_dem=df_dem
    )

    df_pd = convert_plot_data_to_dataframe(dict_for_plots, data_type)

    list_of_keys = list(df_pd.columns)
    list_of_keys.remove("timestamp")
    plots = {}

    if max_days is not None:
        if df_pd["timestamp"].empty:
            logging.warning("The timeseries for {} are empty".format(data_type))
        elif df_pd.timestamp.dtype == np.int64:
            logging.warning(
                "The timeseries for {} do not have correct timestamps, it is likely that you uploaded "
                "a timeseries with more or less values than the number of days multiplied by number of "
                "timesteps within a day.".format(
                    data_type
                )
            )
        else:
            if not isinstance(df_pd["timestamp"], pd.DatetimeIndex):
                dti = dict_values[SIMULATION_SETTINGS][TIME_INDEX]
                df_pd = df_pd.loc[: len(dti) - 1]
                df_pd["timestamp"] = dti
                max_date = df_pd["timestamp"][0] + pd.Timedelta(
                    "{} day".format(max_days)
                )
            df_pd = df_pd.loc[df_pd["timestamp"] < max_date]
        title_addendum = " ({} days)".format(max_days)
    else:
        title_addendum = ""

    for i, component in enumerate(list_of_keys):
        comp_id = component + "-plot"
        fig = create_plotly_line_fig(
            x_data=df_pd["timestamp"],
            y_data=df_pd[component],
            plot_title="{}{}".format(dict_plot_labels[component], title_addendum),
            x_axis_name="Time",
            y_axis_name="kW",
            color_for_plot=get_color(i, color_list),
            file_path=file_path,
        )
        if file_path is None:
            plots[comp_id] = fig

    return plots


def plot_sankey(dict_values):
    """"""
    fig_dict = dict_values[PATHS_TO_PLOTS].get(PLOT_SANKEY, None)
    if fig_dict is not None:
        fig = go.Figure(**fig_dict)
    else:
        fig = go.Figure()
    return fig


def create_plotly_barplot_fig(
    x_data,
    y_data,
    plot_title=None,
    trace_name="",
    legends=None,
    x_axis_name=None,
    y_axis_name=None,
    file_name="barplot.png",
    file_path=None,
):
    r"""
    Create figure for specific capacities barplot

    Parameters
    ----------
    x_data: list, or pandas series
        The list of abscissas of the data required for plotting.

    y_data: list, or pandas series, or list of lists
        The list of ordinates of the data required for plotting.

    plot_title: str
        The title of the plot generated.
        Default: None

    trace_name: str
        Sets the trace name. The trace name appear as the legend item and on hover.
        Default: ""

    legends: list, or pandas series
        The list of the text written within the bars and on hover below the trace_name
        Default: None

    x_axis_name: str
        Default: None

    y_axis_name: str
        Default: None

    file_name: str
        Name of the image file.
        Default: "barplot.png"

    file_path: str
        Path where the image shall be saved if not None

    Returns
    -------
    fig: :class:`plotly.graph_objs.Figure`
        figure object
    """
    fig = go.Figure()

    styling_dict = get_fig_style_dict()
    styling_dict["mirror"] = True

    opts = {}
    if legends is not None:
        opts.update(dict(text=legends, textposition="auto"))

    fig.add_trace(
        go.Bar(
            name=trace_name,
            x=x_data,
            y=y_data,
            marker_color=px.colors.qualitative.D3,
            **opts,
        )
    )

    fig.update_layout(
        xaxis_title=x_axis_name,
        yaxis_title=y_axis_name,
        template="simple_white",
        font_family="sans-serif",
        # TODO use styling dict
        xaxis=go.layout.XAxis(
            showgrid=True,
            gridwidth=1.5,
            zeroline=True,
            mirror=True,
            autorange=True,
            linewidth=1,
            ticks="inside",
            visible=True,
        ),
        yaxis=styling_dict,
        title={
            "text": plot_title,
            "y": 0.90,
            "x": 0.5,
            "font_size": 23,
            "xanchor": "center",
            "yanchor": "top",
        },
        legend_title="Components",
    )

    if file_path is not None:
        # Function call to save the Plotly plot to the disk
        save_plots_to_disk(
            fig_obj=fig,
            file_path=file_path,
            file_name=file_name,
            width=1200,
            height=600,
            scale=5,
        )

    return fig


def plot_optimized_capacities(
    dict_values, file_path=None,
):
    """Plot capacities as a bar chart.

    Parameters
    ----------
    dict_values :
        dict Of all input and output parameters up to F0

    file_path: str
        Path where the image shall be saved if not None
        Default: None

    Returns
    -------
    Dict with html DOM id for the figure as key and :class:`plotly.graph_objs.Figure` as value
    """

    # Add dataframe to hold all the KPIs and optimized additional capacities
    df_capacities = dict_values[KPI][KPI_SCALAR_MATRIX].copy(deep=True)
    df_capacities.drop(
        columns=[TOTAL_FLOW, ANNUAL_TOTAL_FLOW, PEAK_FLOW, AVERAGE_FLOW], inplace=True,
    )
    df_capacities.reset_index(drop=True, inplace=True)

    x_values = []
    y_values = []
    legends = []

    for kpi, cap, unit in zip(
        list(df_capacities[LABEL]),
        list(df_capacities[OPTIMIZED_ADD_CAP]),
        list(df_capacities[UNIT]),
    ):
        if cap > 0:
            x_values.append(kpi)
            y_values.append(cap)
            if unit == "?":
                unit = "kW"
            legends.append("{:.0f} {}".format(cap, unit))

    # Title to add to plot titles
    project_title = ": {}, {}".format(
        dict_values[PROJECT_DATA][PROJECT_NAME],
        dict_values[PROJECT_DATA][SCENARIO_NAME],
    )

    name_file = "optimal_additional_capacities"

    fig = create_plotly_barplot_fig(
        x_data=x_values,
        y_data=y_values,
        plot_title="Optimal additional capacities" + project_title,
        trace_name="capacities",
        legends=legends,
        x_axis_name="Items",
        y_axis_name="Capacities",
        file_name=name_file,
        file_path=file_path,
    )

    return {"capacities_plot": fig}


def create_plotly_flow_fig(
    df_plots_data,
    x_legend=None,
    y_legend=None,
    plot_title=None,
    color_list=None,
    file_name="flows.png",
    file_path=None,
):
    r"""Generate figure of an asset's flow.

    Parameters
    ----------
    df_plots_data: :class:`pandas.DataFrame<frame>`
        dataFrame with timeseries of the asset's energy flow
    x_legend: str
        Default: None

    y_legend: str
        Default: None

    plot_title: str
        Default: None

    color_list: list of str or list to tuple (hexadecimal or rbg code)
        list of colors
        Default: None

    file_name: str
        Name of the image file.
        Default: "flows.png"

    file_path: str
        Path where the image shall be saved if not None
        Default: None

    Returns
    -------
    fig: :class:`plotly.graph_objs.Figure`
        figure object
    """

    fig = go.Figure()
    styling_dict = get_fig_style_dict()
    styling_dict["gridwidth"] = 1.0

    assets_list = list(df_plots_data.columns)
    assets_list.remove("timestamp")

    for i, asset in enumerate(assets_list):
        fig.add_trace(
            go.Scatter(
                x=df_plots_data["timestamp"],
                y=df_plots_data[asset],
                mode="lines",
                line=dict(color=get_color(i, color_list), width=2.5),
                name=asset,
            )
        )

    fig.update_layout(
        xaxis_title=x_legend,
        yaxis_title=y_legend,
        font_family="sans-serif",
        template="simple_white",
        xaxis=styling_dict,
        yaxis=styling_dict,
        title={
            "text": plot_title,
            "y": 0.90,
            "x": 0.5,
            "font_size": 23,
            "xanchor": "center",
            "yanchor": "top",
        },
        legend=dict(y=0.5, traceorder="normal", font=dict(color="black"),),
    )

    if file_path is not None:
        # Function call to save the Plotly plot to the disk
        save_plots_to_disk(
            fig_obj=fig,
            file_path=file_path,
            file_name=file_name,
            width=1200,
            height=600,
            scale=5,
        )

    return fig


def plot_instant_power(dict_values, file_path=None):
    """Plotting timeseries of instantaneous power for each assets within the energy system

    Parameters
    ----------
    dict_values : dict
        all simulation input and output data up to this point

    file_path: str
        Path where the image shall be saved if not None
        Default: None

    Returns
    -------
    multi_plots: dict
       Dict with html DOM id for the figure as keys and :class:`plotly.graph_objs.Figure` as values
    """
    buses_list = list(dict_values[OPTIMIZED_FLOWS].keys())
    multi_plots = {}
    for bus in buses_list:
        df_data = dict_values[OPTIMIZED_FLOWS][bus].copy(deep=True)
        df_data.reset_index(level=0, inplace=True)
        df_data = df_data.rename(columns={"index": "timestamp"})

        # In case SOC of a storage is in df_data the SOC is plotted separately and is
        # removed from the df_data as the plot that shows absolute flows should not
        # contain SOC in %
        if any(SOC in item for item in df_data):
            # if any(SOC in item for item in df_data):
            comp_id = f"{SOC}-{bus}-plot"
            title = (
                bus
                + " storage SOC in LES: "
                + dict_values[PROJECT_DATA][PROJECT_NAME]
                + ", "
                + dict_values[PROJECT_DATA][SCENARIO_NAME]
            )
            # get columns containing SOC and plot SOC
            soc_cols = [s for s in df_data.keys() if SOC in s]
            soc_cols.extend(["timestamp"])
            fig = create_plotly_flow_fig(
                df_plots_data=df_data[soc_cols],
                x_legend="Time",
                y_legend="SOC",
                plot_title=title,
                file_path=file_path,
                file_name=f"SOC_{bus}_power.png",
            )
            if file_path is None:
                multi_plots[comp_id] = fig

            # remove SOC as it is provided in % and does not fit with flow plot
            soc_cols.remove("timestamp")
            df_data.drop(soc_cols, inplace=True, axis=1)

        # create flow plot
        comp_id = f"{bus}-plot"
        title = (
            bus
            + " power in LES: "
            + dict_values[PROJECT_DATA][PROJECT_NAME]
            + ", "
            + dict_values[PROJECT_DATA][SCENARIO_NAME]
        )

        fig = create_plotly_flow_fig(
            df_plots_data=df_data,
            x_legend="Time",
            y_legend=bus + " in kW",
            plot_title=title,
            file_path=file_path,
            file_name=bus + "_power.png",
        )
        if file_path is None:
            multi_plots[comp_id] = fig

    return multi_plots


def create_plotly_piechart_fig(
    title_of_plot,
    names,
    values,
    color_scheme=None,
    file_name="costs.png",
    file_path=None,
):
    r"""Generate figure with piechart plot.

    Parameters
    ----------
    title_of_plot: str
        title of the figure

    names: list
        List containing the labels of the slices in the pie plot.

    values: list
        List containing the values of the labels to be plotted in the pie plot.

    color_scheme: instance of the px.colors class of the Plotly express library
        This parameter holds the color scheme which is palette of colors (list of hex values) to be
        applied to the pie plot to be created.
        Default: None

    file_name: str
        Name of the image file.
        Default: "costs.png"

    file_path: str
        Path where the image shall be saved if not None
        Default: None

    Returns
    -------
    fig: :class:`plotly.graph_objs.Figure`
        figure object
    """

    if color_scheme is None:
        color_scheme = px.colors.qualitative.Set1

    # Wrap the text of the title into next line if it exceeds the length given below
    title_of_plot = textwrap.wrap(title_of_plot, width=75)
    title_of_plot = "<br>".join(title_of_plot)

    fig = go.Figure(
        go.Pie(
            labels=names,
            values=values,
            textposition="inside",
            insidetextorientation="radial",
            texttemplate="%{label} <br>%{percent}",
            marker=dict(colors=color_scheme),
        ),
    )

    fig.update_layout(
        title={
            "text": title_of_plot,
            "y": 0.9,
            "x": 0.5,
            "font_size": 23,
            "xanchor": "center",
            "yanchor": "top",
            "pad": {"r": 5, "l": 5, "b": 5, "t": 5},
        },
        font_family="sans-serif",
        height=500,
        width=700,
        autosize=True,
        legend=dict(orientation="v", y=0.5, yanchor="middle", x=0.95, xanchor="right",),
        margin=dict(l=10, r=10, b=50, pad=2),
        uniformtext_minsize=18,
    )
    fig.update_traces(hoverinfo="label+percent", textinfo="label", textfont_size=18)

    if file_path is not None:
        # Function call to save the Plotly plot to the disk
        save_plots_to_disk(
            fig_obj=fig,
            file_path=file_path,
            file_name=file_name,
            width=1200,
            height=600,
            scale=5,
        )

    return fig


def plot_piecharts_of_costs(dict_values, file_path=None):
    """Plotting piecharts of different cost parameters (ie. annuity, total cost, etc...)

    Parameters
    ----------
    dict_values : dict
        all simulation input and output data up to this point

    file_path: str
        Path where the image shall be saved if not None
        Default: None

    Returns
    -------
    pie_plots: dict
       Dict with html DOM id for the figure as keys and :class:`plotly.graph_objs.Figure` as values
    """

    df_pie_data = convert_costs_to_dataframe(dict_values)

    # Initialize an empty list and a dict for use later in the function
    pie_plots = {}
    pie_data_dict = {}

    # df_pie_data.reset_index(drop=True, inplace=True)
    columns_list = list(df_pie_data.columns)
    columns_list.remove(LABEL)

    # Iterate through the list of columns of the DF which are the KPIs to be plotted
    for kp_indic in columns_list:

        # Assign an id for the plot
        comp_id = kp_indic + "plot"

        kpi_part = ""

        # Make a copy of the DF to make various manipulations for the pie chart plotting
        df_temp = df_pie_data.copy()

        # Get the total value for each KPI to use in the title of the respective pie chart
        df_temp2 = df_temp.copy()
        df_temp2.set_index(LABEL, inplace=True)
        total_for_title = df_temp2.at["Total", kp_indic]

        # Drop the total row in the dataframe
        df_temp.drop(df_temp.tail(1).index, inplace=True)
        # Gather the data for each asset for the particular KPI, in a dict
        for row_index in df_temp.index:
            pie_data_dict[df_temp.at[row_index, LABEL]] = df_temp.at[
                row_index, kp_indic
            ]

        # Remove negative values (such as the feed-in sinks) from the dict
        pie_data_dict = {k: v for (k, v) in pie_data_dict.items() if v > 0}

        # Get the names and values for the pie chart from the above dict
        names_plot = list(pie_data_dict.keys())
        values_plot = list(pie_data_dict.values())

        # Below loop determines the first part of the plot title, according to the kpi being plotted
        if "annuity" in kp_indic:
            kpi_part = "Annuity Costs ("
            file_name = "annuity"
            scheme_choosen = px.colors.qualitative.Set1
        elif "investment" in kp_indic:
            kpi_part = "Upfront Investment Costs ("
            file_name = "upfront_investment_costs"
            scheme_choosen = px.colors.diverging.BrBG
        elif "om" in kp_indic:
            kpi_part = "Operation and Maintenance Costs ("
            file_name = "operation_and_maintainance_costs"
            scheme_choosen = px.colors.sequential.RdBu

        # Title to add to plot titles
        project_title = ": {}, {}".format(
            dict_values[PROJECT_DATA][PROJECT_NAME],
            dict_values[PROJECT_DATA][SCENARIO_NAME],
        )

        # Title of the pie plot
        plot_title = (
            kpi_part
            + str(round(total_for_title))
            + " "
            + dict_values[ECONOMIC_DATA][CURR]
            + ") "
            + project_title
        )

        fig = create_plotly_piechart_fig(
            title_of_plot=plot_title,
            names=names_plot,
            values=values_plot,
            color_scheme=scheme_choosen,
            file_name=file_name,
            file_path=file_path,
        )

        if file_path is None:
            pie_plots[comp_id] = fig

    return pie_plots



"""
Module E4 - Verification of results
===================================

- Detect excessive excess generation
- Verify that minimal renewable share constraint is adhered to
- Verfiy that maximum emission constraint is adhered to
- Verfiy that net zero energy (NZE) constraint is adhered to

"""

import logging
import pandas as pd
from multi_vector_simulator.utils.constants_json_strings import (
    CONSTRAINTS,
    MINIMAL_RENEWABLE_FACTOR,
    VALUE,
    KPI,
    KPI_SCALARS_DICT,
    RENEWABLE_FACTOR,
    OPTIMIZED_FLOWS,
    MAXIMUM_EMISSIONS,
    TOTAL_EMISSIONS,
    ENERGY_STORAGE,
    TIMESERIES_SOC,
    NET_ZERO_ENERGY,
    DEGREE_OF_NZE,
)


def minimal_constraint_test(dict_values, minimal_constraint, bounded_result):
    r"""
    Test if minimal constraint was correctly applied

    Parameters
    ----------
    dict_values: dict
        Dict of all simulation information

    minimal_constraint: str
        Key to access the value of the minimal bound of parameter subjected to constraint to be tested

    bounded_result: str
        Key to access the value of the resulting parameter to be compared to minimal_bound

    Returns
    -------
    Nothing

    - Nothing if the constraint is confirmed
    - Prints logging.warning message if the deviation from the constraint is < 10^-6.
    - Prints a logging.error message if the deviation from the constraint is >= 10^-6.

    Notes
    -----
    Executed to test
    - MINIMAL_DEGREE_OF_AUTONOMY  vs. RENEWABLE_FACTOR
    - MINIMAL_RENEWABLE_FACTOR vs. DEGREE_OF_AUTONOMY

    Tested with:
    - E4.test_minimal_constraint_test_passes()
    - E4.test_minimal_constraint_test_fails()

    """
    if dict_values[CONSTRAINTS][minimal_constraint][VALUE] > 0:
        boolean_test = (
            dict_values[KPI][KPI_SCALARS_DICT][bounded_result]
            >= dict_values[CONSTRAINTS][minimal_constraint][VALUE]
        )
        if boolean_test is False:
            deviation = (
                dict_values[CONSTRAINTS][minimal_constraint][VALUE]
                - dict_values[KPI][KPI_SCALARS_DICT][bounded_result]
            ) / dict_values[CONSTRAINTS][minimal_constraint][VALUE]
            if abs(deviation) < 10 ** (-6):
                logging.warning(
                    f"{minimal_constraint} constraint strictly not fulfilled, but deviation is less then e6."
                )
            else:
                logging.error(
                    f"ATTENTION: {minimal_constraint} constraint NOT fulfilled! The deviation is {round(deviation, 5)}."
                )
                return False

        else:
            logging.debug(
                f"{minimal_constraint} of {dict_values[CONSTRAINTS][MINIMAL_RENEWABLE_FACTOR][VALUE]} is fullfilled."
            )


def maximum_emissions_test(dict_values):
    r"""
    Tests if maximum emissions constraint was correctly applied.

    Parameters
    ----------
    dict_values : dict
        all input parameters and results up to E0

    Returns
    -------
    Nothing

    - Nothing if the constraint is confirmed
    - Prints logging.warning message if the difference from the constraint is < 10^-6.
    - Prints a logging.error message if the difference from the constraint is >= 10^-6.

    Notes
    -----
    Tested with:
    - E4.test_maximum_emissions_test_passes()
    - E4.test_maximum_emissions_test_fails()

    """
    if dict_values[CONSTRAINTS][MAXIMUM_EMISSIONS][VALUE] is not None:
        boolean_test = (
            dict_values[KPI][KPI_SCALARS_DICT][TOTAL_EMISSIONS]
            <= dict_values[CONSTRAINTS][MAXIMUM_EMISSIONS][VALUE]
        )
        if boolean_test is False:
            diff = (
                dict_values[CONSTRAINTS][MAXIMUM_EMISSIONS][VALUE]
                - dict_values[KPI][KPI_SCALARS_DICT][TOTAL_EMISSIONS]
            )
            if abs(diff) < 10 ** (-6):
                logging.warning(
                    "Maximum emissions criterion strictly not fulfilled, but difference is less then e6."
                )
            else:
                logging.error(
                    f"ATTENTION: Maximum emissions factor criterion NOT fulfilled! The difference is {round(diff, 6)}."
                )
                return False

        else:
            logging.debug(
                f"Maximum emissions constraint of {dict_values[CONSTRAINTS][MAXIMUM_EMISSIONS][VALUE]} is fulfilled."
            )


def net_zero_energy_constraint_test(dict_values):
    r"""
    Tests if net zero energy constraint was correctly applied.

    Parameters
    ----------
    dict_values : dict
        all input parameters and results up to E0

    Returns
    -------
    Nothing

    - Nothing if the constraint is confirmed
    - Prints logging.warning message if the difference from the constraint is < 10^-6.
    - Prints a logging.error message if the difference from the constraint is >= 10^-6.

    Notes
    -----
    Tested with:
    - E4.test_net_zero_energy_constraint_test_fails()
    - E4.test_net_zero_energy_constraint_test_passes()

    """
    if dict_values[CONSTRAINTS][NET_ZERO_ENERGY][VALUE] is True:
        boolean_test = dict_values[KPI][KPI_SCALARS_DICT][DEGREE_OF_NZE] >= 1
        if boolean_test is False:
            diff = 1 - dict_values[KPI][KPI_SCALARS_DICT][DEGREE_OF_NZE]
            if abs(diff) <= 10 ** (-6):
                logging.warning(
                    "Net zero energy criterion strictly not fulfilled, but difference is less then e6."
                )
            else:
                logging.error(
                    f"ATTENTION: Net zero energy criterion NOT fulfilled! The difference is {round(diff, 6)}."
                )
                return False

        else:
            logging.debug(
                f"Net zero energy constraint of is fulfilled with a degree of NZE of {dict_values[KPI][KPI_SCALARS_DICT][DEGREE_OF_NZE]}."
            )


def detect_excessive_excess_generation_in_bus(dict_values):
    r"""
    Warning for any bus with excessive excess generation is given.

    A logging.warning message is printed when the ratio between total outflows and
    total inflows of a bus is < 0.9.

    Parameters
    ----------
    dict_values

    Returns
    -------
    - Nothing if the there is no excessive excess generation
    - Prints logging.warning message for every bus with excessive excess generation.

    """
    for bus_label, bus_df in dict_values[OPTIMIZED_FLOWS].items():
        # disregard excess sinks
        cols = [col for col in bus_df.columns if "excess" not in col]
        df = pd.DataFrame(bus_df[cols].sum(), columns=["sum"])
        # get total in- and outflow of bus
        total_inflow_bus = df[df > 0].dropna().sum().values[0]
        total_outflow_bus = abs(df[df < 0].dropna().sum().values[0])
        if not total_inflow_bus == 0:
            # give a warning in case ratio > 0.9
            ratio = total_outflow_bus / total_inflow_bus
            if ratio < 0.9:
                msg = f"Attention, on bus {bus_label} there is excessive excess generation, totalling up to {round((1 - ratio) * 100)}% of the inflows. The total inflows are {round(total_inflow_bus)} and outflows {round(total_outflow_bus)}  It seems to be cheaper to have this excess generation than to install more capacities that forward the energy carrier to other busses (if those assets can be optimized)."
                logging.warning(msg)


def verify_state_of_charge(dict_values):
    r"""
    This function checks the state of charge of each storage component
    It raises warning log messages if the SoC has a physically infeasible value

    Parameters
    ----------
    dict_values: dict
        Dictionary with all information regarding the simulation, specifically including the energyStorage assets

    Returns
    -------
    - Nothing if there are no physically infeasible SoC values for the storage components
    - Prints log messages to console and log file if there are physically impossible SoC values

    Notes
    -----
    Tested with:
    - test_E4_verification.test_verify_state_of_charge_feasible()
    - test_E4_verification.test_verify_state_of_charge_soc_below_zero()
    - test_E4_verification.test_verify_state_of_charge_soc_above_zero()

    """
    # Dict holding the data of all of the storage components
    storage_data_dict = dict_values[ENERGY_STORAGE]

    # Check if storage components are present in the energy system and only then verify SoC values
    if bool(storage_data_dict) is not False:
        # Loop through the storage components
        for storage in list(storage_data_dict.keys()):
            # Get the SoC time-series in a pandas series data structure
            soc_timeseries = storage_data_dict[storage][TIMESERIES_SOC]
            # Check the SoC time-series for abnormal values
            if (soc_timeseries > 1).any():
                logging.warning(
                    f"SoC of {storage} has at least one time step where its value is greater than 1. This is a physically impossible value!"
                )
            elif (soc_timeseries < 0).any():
                logging.warning(
                    f"SoC of {storage} has at least one time step where its value is less than 0. This is a physically impossible value!"
                )
